# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

import numpy as np
from scipy.stats import chi2

from pathpy.utils import Log, Severity
from pathpy.utils.exceptions import PathpyError, PathsTooShort, PathpyNotImplemented
from pathpy.classes.higher_order_network import HigherOrderNetwork
from pathpy.classes.paths import Paths


np.seterr(all='warn')


class MultiOrderModel:
    """Instances of this class represent a hierarchy of
        higher-order networks which collectively represent
        a multi-order model for path statistics. """

    def __init__(self, paths, max_order=1):
        """Generates a hierarchy of higher-order models for the given path statistics,
        up to a given maximum order

        Parameters
        ----------
        paths: Paths
            the paths instance for which the model should be created
        max_order: int
            the maximum order of the multi-order model
        """
        assert paths.max_subpath_length >= max_order, \
            'Error: Construction of multi-order model with maximum order M ' \
            'requires sub path statistics up to length M'

        # the paths object from which this multi-order model was created
        self.paths = paths

        # A dictionary containing the layers of HigherOrderNetworks, where
        # layers[k] contains the network of order k
        self.layers = {}
        
        # a dictionary of transition matrices for all layers of the model
        self.transition_matrices = {}

        self.add_layers(max_order)

    @property
    def max_order(self):
        """the current maximum order of the multi-order model"""
        orders = list(self.layers.keys())
        if not orders:
            return None
        else:
            return max(orders)

    def __add_layers_parallel(self, orders):
        paths = self.paths
        try:
            import pathos as _pa
        except ImportError:  # pragma: no cover
            raise ImportError('Error while importing module "pathos",'
                              'to use the parallel feature "pathos" '
                              'needs to be installed')

        def parallel(order_k):  # pragma: no cover
            Log.add('Generating ' + str(order_k) + '-th order network layer ...')
            p_layer = HigherOrderNetwork(paths, k=order_k, null_model=False)

            # compute transition matrices for all layers. In order to use the
            # maximally available statistics, we always use sub paths in the
            # calculation
            trans_mat = p_layer.transition_matrix(include_subpaths=True)

            Log.add('... finished')
            return [order_k, p_layer, trans_mat]

        pool = _pa.multiprocessing.ProcessPool()
        results = pool.map(parallel, orders)

        # save results
        for k, layer, transition_mat in results:
            self.layers[k] = layer
            self.transition_matrices[k] = transition_mat

    def __add_layers_sequential(self, orders):
        paths = self.paths

        for k in sorted(orders):
            Log.add('Generating %d-th order layer ...' % k)
            self.layers[k] = HigherOrderNetwork(paths, k, null_model=False)

            # compute transition matrices for all layers. In order to use the
            # maximally available statistics, we always use sub paths in the
            # calculation
            self.transition_matrices[k] = self.layers[k].transition_matrix(include_subpaths=True)

        Log.add('finished.')

    def add_layers(self, max_order):
        """Add higher order layers

        Parameters
        ----------
        max_order: int
            up to which order to add higher order layers, if below the current maximum the
            operation will have no effect and the HigherOrderNetwork will remain unchanged.

        """
        from pathpy import ENABLE_MULTICORE_SUPPORT

        current_max_order = self.max_order if self.max_order else -1
        if max_order < 0:
            raise PathpyError("max_order must be a positive integer not %d" % max_order)

        if max_order <= current_max_order:
            Log.add("Layers up to order %d already added. Nothing changed." % self.max_order)

        orders_to_add = list(range(current_max_order+1, max_order+1))
        if len(orders_to_add) > 1 and ENABLE_MULTICORE_SUPPORT:
            self.__add_layers_parallel(orders_to_add)
        else:
            self.__add_layers_sequential(orders_to_add)

    def summary(self):
        """
        Returns a string containing basic summary information
        of this multi-order model instance.
        """
        summary_fmt = (
            "Multi-order model (max. order = {order}, "
            "DoF (paths/ngrams) = {dof_path} / {dof_ngram})\n"
            '==========================================================================\n'
        )
        summary = summary_fmt.format(
            order=self.max_order,
            dof_path=self.degrees_of_freedom(assumption='paths'),
            dof_ngram=self.degrees_of_freedom(assumption='ngrams')
        )
        layer_fmt = ("Layer k = {k} \t {vcount} nodes, {ecount} links, {sum_path} paths, "
                     "DoF (paths/ngrams) = {dof_paths} / {dof_ngram} \n")

        for k in range(self.max_order + 1):
            vcount = self.layers[k].vcount()
            ecount = self.layers[k].ecount()
            sum_path = self.layers[k].total_edge_weight().sum()
            dof_paths = int(self.layers[k].degrees_of_freedom('paths'))
            dof_ngram = int(self.layers[k].degrees_of_freedom('ngrams'))

            layer_sum = layer_fmt.format(k=k, vcount=vcount, ecount=ecount,
                                         sum_path=sum_path, dof_paths=dof_paths,
                                         dof_ngram=dof_ngram)
            summary += layer_sum
        return summary

    def save_state_file(self, filename, layer=None, infomap_indexing=None):
        """Saves the multi-order model in state file format suitable to be used with
         InfoMap

        Parameters
        ----------
        filename
        layer: int
            if none, all layers will be export. If set to k, only the k-th layer of the
            model will be exported.
        infomap_indexing: dict
            if none, standard pathpy indices will be used in the export of state files.
            This can be set to a custom index dictionary in which infomap_indexing[k]
            contains a dictionary that maps k-th order nodes to a custom node index.
            This is useful to create state files with consistent indices from multiple
            MultiOrderModels

        Returns
        -------

        """
        assert layer, 'Export of all layers is currently not supported'

        name_map = self.layers[layer].node_to_name_map()
        first_layer_map = self.layers[1].node_to_name_map()

        trans_mat = self.layers[layer].transition_matrix()

        file = open(filename, 'w')

        file.write('# this file was generated by pathpy\n')

        # Note: InfoMap requires consecutive indexing of nodes!
        if infomap_indexing:
            file.write('*Vertices {0}\n'.format(len(infomap_indexing[1])))
            # sort indices
            reverse_index = {}
            for name, node_idx in infomap_indexing[1].items():
                reverse_index[node_idx] = name
            sorted_indices = list(reverse_index.keys())
            for node_idx in sorted_indices:
                file.write('{0} "{1}"\n'.format(node_idx, reverse_index[node_idx]))
        else:
            file.write('*Vertices {0}\n'.format(self.layers[1].vcount()))
            for i in self.layers[1].nodes:
                first = first_layer_map[i]
                second = self.layers[1].nodes[i]

                file.write('{0} "{1}"\n'.format(first, second))

        # Write higher-order nodes to states section
        file.write('*States {0}\n'.format(self.layers[layer].vcount()))
        for v in self.layers[layer].nodes:
            if infomap_indexing:
                v_ix = infomap_indexing[layer][v]
            else:
                v_ix = name_map[v]
            v_path = self.layers[layer].higher_order_node_to_path(v)

            # each line contains uniqueID physicalID [name]
            if infomap_indexing:
                file.write(
                    '{0} {1} "{2}"\n'.format(v_ix, infomap_indexing[1][v_path[-1]], v))
            else:
                file.write(
                    '{0} {1} "{2}"\n'.format(v_ix, first_layer_map[v_path[-1]], v))

        file.write('*Links {0}\n'.format(self.layers[layer].ecount()))
        for e in self.layers[layer].edges:
            source = e[0]
            target = e[1]

            # Get source and target paths
            # source_p = self.layers[layer].higher_order_node_to_path(source)
            # source_t = self.layers[layer].higher_order_node_to_path(target)

            source_ix = name_map[source]
            target_ix = name_map[target]

            # Get edge weight
            # w_st = self.layers[layer].edges[e][1]

            # Get transition probability
            trans_prop = trans_mat[target_ix, source_ix]

            # Write entry to file
            # each line contains from to [weight]
            if infomap_indexing:
                idx_s = infomap_indexing[layer][source]
                idx_t = infomap_indexing[layer][target]
                file.write('{} {} {}\n'.format(idx_s, idx_t, trans_prop))
            else:
                file.write('{} {} {}\n'.format(source_ix, target_ix, trans_prop))

        file.close()

    def __str__(self):
        """
        Returns the default string representation of
        this multi-order model instance
        """
        return self.summary()

    def likelihood(self, paths=None, max_order=None, log=True):
        """Calculates the likelihood of a multi-order
        network model up to a maximum order max_order based on all
        path statistics.

        Parameters
        ----------
        paths:
            the path statistics to be used in the likelihood calculation
        max_order:
            the maximum layer order to take into account for the likelihood calculation.
            For the default value None, all orders will be used for the likelihood
            calculation.
        log: bool
            Whether or not to return the log likelihood (default: True)

        Returns
        -------

        """
        max_order = self.max_order if max_order is None else max_order
        assert max_order <= self.max_order, \
            'Error: max_order cannot be larger than maximum order of multi-order network'

        # add log-likelihoods of multiple model layers,
        # assuming that paths are independent
        likelihood = np.float64(0)

        for k in range(0, max_order + 1):
            if k < max_order:
                p = self.layer_likelihood(paths, k, consider_longer_paths=False, log=True)[0]
            else:
                p = self.layer_likelihood(paths, k, consider_longer_paths=True, log=True)[0]
            # print('Log L(k=' + str(k) + ') = ' + str(p))
            assert p <= 0, 'Layer Log-Likelihood out of bounds'
            likelihood += p
        assert likelihood <= 0, 'Log-Likelihood out of bounds'

        return likelihood if log else np.exp(likelihood)

    @staticmethod
    def factorial(n, log=True):  # pragma: no cover
        """

        Parameters
        ----------
        n: int
            computes factorial of n
        log: bool
            whether or not to return the (natural) logarithm of the factorial

        Returns
        -------
        float
        """

        f = np.float64(0)
        n_ = np.float64(n)
        if n > 20:  # use Stirling's approximation
            try:
                f = (n_ * np.log(n_) - n_ + 0.5 * np.log(2.0 * np.pi * n_)
                     + 1.0 / (12.0 * n_) - 1 / (360.0 * n_ ** 3.0))
            except Warning as w:
                msg = 'Factorial calculation for n={}: {}'.format(n, w)
                Log.add(msg, severity=Severity.WARNING)

        else:
            f = np.log(np.math.factorial(n))

        if log:
            return f
        else:
            return np.exp(f)

    def layer_likelihood(self, paths=None, l=1, consider_longer_paths=True, log=True,
                         min_path_length=None):
        """
        Calculates the (log-)likelihood of the **first** l layers of a multi-order
        network model using all observed paths of (at least) length l

        Parameters
        ----------
        paths: Paths
            the path statistics for which to calculate the layer likelihood
        l: int
            number of layers for which likelihood shall be calculated Paths of length l
            (and possibly longer) will be used to calculate the likelihood of model layers
            for all orders up to l
        consider_longer_paths: bool
            whether or not to include paths longer than l in the calculation of the
            likelihood. In general, when calculating the likelihood of a multi-order model
            which combines orders from 1 to l, this should be set to true only for the
            value of l that corresponds to the largest order in the model.
        log: bool
             whether to compute Log-Likelihood (default: True)
        min_path_length: int
            minimum length of paths which enter the likelihood calculation. For the
            default value None, all paths with at least length l will be considered.

        Returns
        -------
        float
            the (log-)likelihood of the model layer given the path statistics
        """
        # m is the maximum length of any path in the data
        if paths is None:
            paths = self.paths
        max_len_obs = max(paths.paths)

        assert max_len_obs >= l and len(paths.paths[l]) > 0, \
            'Error: there are no paths of length l or longer'

        if min_path_length is None:
            min_path_length = l

        # Set maximum length of paths to consider in likelihood calculation
        if consider_longer_paths:
            maxL = max_len_obs
        else:
            maxL = l

        # create index maps to map node names to matrix indices
        indexmaps = {}
        for k in range(0, l+1):
            indexmaps[k] = self.layers[k].node_to_name_map()

        # For the paths S_k of length k (or longer) that we observe, we need to calculate
        # the probability of observing all paths in S_k based on the probabilities of
        # individual paths (which are calculated using the underlying Markov model(s))

        # n is the total number of path observations
        n = 0

        # Initialize likelihood
        likelihood = 0

        # compute likelihood for all longest paths
        # up to the maximum path length maxL
        for k in range(min_path_length, maxL + 1):
            for p in paths.paths[k]:
                # Only consider observations as *longest* path
                freq = paths.paths[k][p][1]
                if freq > 0:
                    n += freq  # Add m_i observations of path p to total number of observations n
                    likelihood += self.path_likelihood(p, freq, l, log=True, index_maps=indexmaps)
            if n == 0:
                likelihood = 0
        if log:
            assert likelihood <= 0, 'Log-Likelihood out of bounds'
            return likelihood, n
        else:
            assert 0 <= likelihood <= 1, 'Likelihood out of bounds'
            return np.exp(likelihood), n

    def path_likelihood(self, path, freq=1, layer=1, log=True, index_maps=None):
        """compute the path likelihood for a single path based on the observed transitions

        Parameters
        ----------
        path: tuple
            the path for which the likelihood should be computed, the path must be a tuple whose
            elements represent the path, i.e. ('a', 'b', 'd')
        freq: int
            the frequency of the path
        layer: int
            the layer up to which the likelihood should be computed
        log: bool
            true if the log-likelihood should be returned
        index_maps: dict
            a dictionary mapping the nodes for all orders to their index in the transition matrix.
            For simple calls this is computed automatically, however if the a lot of paths will be
            calculated it is best to pre-compute it and pass it.

        Returns
        -------
        float:
            likelihood or log-likelihood of path

        Raises
        -----
        PathpyNotImplemented

        Examples
        -------
        >>> layer = 2
        >>> p = Paths()
        >>> p.add_path_tuple(('1', '3', '2'))
        >>> p.add_path_tuple(('3', '2', '1'))
        >>> p.add_path_tuple(('1', '2', '1'))
        >>> mom = MultiOrderModel(p, max_order=layer)
        >>> # a precomputed index map can be obtained as follows
        >>> index_maps = {k: mom.layers[k].node_to_name_map() for k in range(0, layer+1)}
        >>> format(mom.path_likelihood(('1', '2', '1'), log=False, layer=layer), '.2f')
        '0.22'
        >>> # trying to compute the likelihood of an nonexistent path
        >>> format(mom.path_likelihood(('1', '2', '2'), log=False, layer=layer), '.2f')
        Traceback (most recent call last):
        ...
        pathpy.utils.exceptions.PathpyNotImplemented: The path segment '(2,2)' has not been \
observed and therefore the likelihood cannot be computed.


        """
        if index_maps is None:
            index_maps = {k: self.layers[k].node_to_name_map() for k in range(0, layer+1)}

        likelihood = 0
        # special case: to calculate the likelihood of the path based on a
        # zero-order model we use the 'start' -> v transitions in the
        # respective model instance
        if layer == 0:
            try:
                for s in range(len(path)):
                    source = index_maps[0]['start']
                    target = index_maps[0][path[s]]
                    likelihood += np.log(self.transition_matrices[0][target, source]) * freq
            except KeyError as e:
                msg = ("The path segment '({})' has not been observed and therefore the "
                       "likelihood cannot be computed.").format(e.args[0])
                raise PathpyNotImplemented(msg)

        # general case: compute likelihood of path based on hierarchy of higher-order models
        else:
            # 1.) transform the path into a sequence of (two or more) l-th-order nodes
            nodes = self.layers[layer].path_to_higher_order_nodes(path)
            # print('l-th order path = ', str(nodes))

            # 2.) nodes[0] is the prefix of the k-th order transitions,  which we can transform
            # into multiple transitions in lower order models. Example: for a path a-b-c-d of
            # length three, the node sequence at order l=3 is ['a-b-c', 'b-c-d'] and thus the
            # prefix is 'a-b-c'.
            prefix = nodes[0]

            # 3.) We extract the transitions for the prefix based on models of orders k_<l. In
            # our example, we have the transitions ... (a-b, b-c) for k_=2 (a, b) for k_=1,
            # and (start, a) for k_=0
            transitions = {}

            # for all k_<l in descending order
            for k_ in range(layer - 1, -1, -1):
                x = prefix.split(self.layers[k_].separator)
                transitions[k_] = self.layers[k_].path_to_higher_order_nodes(x)
                prefix = transitions[k_][0]

            # 4.) Using Bayes theorem, we calculate the likelihood of a path a-b-c-d-e of length
            # four for l=4 as a single transition in a fourth-order model, and four additional
            # transitions in the k_=0, 1, 2 and 3-order models, i.e. we have ... P(a-b-c-d-e) =
            # P(e|a-b-c-d) * [ P(d|a-b-c) * P(c|a-b) * P(b|a) * P(a) ] If we were to model the
            # same path based on model hierarchy with a maximum order of l=2, we instead have
            # three transitions in the second-order model and two additional transitions in the
            # k_=0 and k_=1 order models for the prefix 'a-b' ... P(a-b-c-d-e) = P(e|c-d) * P(
            # d|b-c) * P(c|a-b) * [ P(b|a) * P(a) ]

            # First multiply the transitions in the l-th order model ...
            transition_matrix = self.transition_matrices[layer]
            try:
                for s in range(len(nodes)-1):
                    idx_s1 = index_maps[layer][nodes[s + 1]]
                    idx_s0 = index_maps[layer][nodes[s]]
                    trans_mat = transition_matrix[idx_s1, idx_s0]
                    likelihood += np.log(trans_mat) * freq
                # ... then multiply additional transition probabilities for the prefix ...
                for k_ in range(layer):
                    trans_idx0 = index_maps[k_][transitions[k_][0]]
                    trans_idx1 = index_maps[k_][transitions[k_][1]]
                    trans_mat = self.transition_matrices[k_]
                    likelihood += np.log(trans_mat[trans_idx1, trans_idx0]) * freq
            except KeyError as e:
                msg = ("The path segment '({})' has not been observed and therefore the "
                       "likelihood cannot be computed.").format(e.args[0])
                raise PathpyNotImplemented(msg)

        if log:
            return likelihood
        else:
            return np.exp(likelihood)

    def degrees_of_freedom(self, max_order=None, assumption="paths"):
        """
        Calculates the degrees of freedom of the model based on
        different assumptions, and taking into account layers up to
        a maximum order.

        Parameters
        ----------
        max_order: int
            the maximum order up to which model layers shall be taken into account
        assumption: str
            if set to 'paths', for the degree of freedom calculation only paths in the
            first-order network topology will be considered. This is needed whenever we
            model paths in a *given* network topology. If set to 'ngrams' all possible
            n-grams will be considered, independent of whether they are valid paths in the
            first-order network or not. The 'ngrams' and the 'paths' assumption coincide
            if the first-order network is fully connected, i.e. if all possible paths
            actually occur.

        Returns
        -------

        """
        if max_order is None:
            max_order = self.max_order
        assert max_order <= self.max_order, \
            'Error: max_order cannot be larger than maximum order of multi-order network'

        dof = 0

        # Sum degrees of freedom of all model layers up to max_order
        for i in range(0, max_order + 1):
            dof += self.layers[i].degrees_of_freedom(assumption)

        return int(dof)

    def model_size(self, max_order):
        """
        Returns the total number of non-zero
        transition matrix entries in all
        model layers
        """
        max_order = self.max_order if max_order is None else max_order
        assert max_order <= self.max_order, \
            'Error: max_order cannot be larger than maximum order of multi-order network'

        size = 0
        for i in range(0, max_order + 1):
            size += self.layers[i].model_size()
        return int(size)

    def likelihood_ratio_test(self, paths=None, max_order_null=0, max_order=1,
                              assumption='paths', significance_threshold=0.01):
        """
        Performs a likelihood-ratio test between two multi-order models with given
        maximum orders, where maxOrderNull serves as null hypothesis and max_order
        serves as alternative hypothesis. The null hypothesis is rejected if the
        p-value for the observed paths under the null hypothesis is smaller than the
        given significance threshold.

        Applying this test makes the assumption that we have nested models, i.e. that
        the null model is contained as a special case in the parameter space of the
        more complex model. If we assume that the path constraint holds, this is not
        true for the test of the first- against the zero-order model (since some
        sequences of the zero order model cannot be generated in the first-order
        model). However, since the set of possible higher-order transitions is
        generated based on the first-order model, the nestedness property holds for all
        higher order models.

        Parameters
        ----------
        paths:
            the path data to be used in the likelihood ratio test
        max_order_null:
            maximum order of the multi-order model to be used as a null hypothesis
        max_order: int
            maximum order of the multi-order model to be used as alternative hypothesis
        assumption: str
            paths or ngrams
        significance_threshold: float
            the threshold for the p-value below which to accept the alternative hypothesis

        Returns
        -------
        tuple
            a tuple of the format (reject, p) which captures whether or not the null
            hypothesis is rejected in favor of the alternative hypothesis,
            as well as the p-value that led to the decision
        """
        assert max_order_null < max_order, \
            'Error: order of null hypothesis must be smaller than order of ' \
            'alternative hypothesis'
        # let L0 be the likelihood for the null model and L1 be the likelihood for the
        # alternative model

        # we first compute a test statistic x = -2 * log (L0/L1) = -2 * (log L0 - log L1)
        x = -2 * (self.likelihood(paths, max_order=max_order_null, log=True) -
                  self.likelihood(paths, max_order=max_order, log=True))

        # we calculate the additional degrees of freedom in the alternative model
        dof_diff = (
            self.degrees_of_freedom(max_order=max_order, assumption=assumption) -
            self.degrees_of_freedom(max_order=max_order_null, assumption=assumption)
        )

        Log.add('Likelihood ratio test for K_opt = ' + str(max_order) + ', x = ' + str(x))
        Log.add('Likelihood ratio test, d_1-d_0 = ' + str(dof_diff))

        # if the p-value is *below* the significance threshold, we reject the null
        # hypothesis
        p = 1 - chi2.cdf(x, dof_diff)

        Log.add('Likelihood ratio test, p = ' + str(p))
        return (p < significance_threshold), p

    def estimate_order(self, paths=None, stop_at_order=100, significance_threshold=0.01):
        """Selects the optimal maximum order of a multi-order network model for the
        observed paths, based on a likelihood ratio test with p-value threshold of p

        By default, all orders up to the maximum order of the multi-order model will be
        tested.

        Parameters
        ----------
        paths: Paths
             The path statistics for which to perform the order selection.
             It defaults to the path statistics the MultiOrderModel was created from.
        stop_at_order: int
            The maximum order up to which the multi-order model shall be tested.
            If the test fails earlier the remaining orders will not be computed.
            The default is 100
        significance_threshold: float
            the threshold for the p-value below which to accept the alternative hypothesis

        Returns
        -------
        int

        """
        import warnings
        assert stop_at_order > 1, 'Error: stop_at_order must be larger than one'

        # Test for highest order that passes, likelihood ratio test against null model
        max_accepted_order = 1
        for k in range(2, stop_at_order + 1):

            if k >= self.max_order:
                try:
                    self.add_layers(k)
                except PathsTooShort:
                    msg = ("optimal order is at least %d, but could be higher. Paths are too short"
                           "to create higher orders layers." % max_accepted_order)
                    warnings.warn(msg)
                    break

            accept, p_value = self.likelihood_ratio_test(
                paths, max_order_null=k - 1, max_order=k,
                significance_threshold=significance_threshold
            )
            if accept:
                max_accepted_order = k
            else:
                break

        if stop_at_order == max_accepted_order:
            msg = ("order is at least %d, but could be higher, "
                   "try to increase `stop_at_order`" % stop_at_order)
            warnings.warn(msg)
        return max_accepted_order

    def test_network_hypothesis(self, paths, method='AIC'):
        """
        Tests whether the assumption that paths are constrained
        to the (first-order) network topology is justified.
        Roughly speaking, this test yields true if the gain in
        explanatory power that is due to the network topology
        justifies the additional model complexity.

        The decision will be made based on a comparison between the zero-
        and the first-order layer of the model. Different from the multi-order
        model selection method implemented in estimate_order and likelihoodRatioTest,
        here we do *not* consider nested models, so we cannot use a likelihood ratio
        test. We instead use the AIC or BIC.
        """
        from pathpy.utils.exceptions import PathpyError
        assert method in ['AIC', 'BIC', 'AICc'], \
            'Expected method AIC, AICc or BIC "%s" given.' % method

        # count number of omitted paths with length zero
        p_sum = 0
        for p in paths.paths[0]:
            p_sum += paths.paths[0][p][1]
        if p_sum > 0:
            msg = 'Omitting {} zero-length paths ' \
                  'for test of network assumption'.format(p_sum)
            Log.add(msg, Severity.INFO)

        # log-likelihood and observation count of zero-order model
        likelihood_0, n_0 = self.layer_likelihood(paths, l=0, consider_longer_paths=True,
                                                  log=True, min_path_length=1)

        # log-likelihood and observation count of first-order model
        likelihood_1, n_1 = self.layer_likelihood(paths, l=1, consider_longer_paths=True,
                                                  log=True, min_path_length=1)

        # By definition, the number of observations for both models should be the total
        # weighted degree of the first-order network
        if n_0 != n_1:
            raise PathpyError(
                'Observation count for 0-order ({n0}) and '
                '1-st order model ({n1}) do not match'.format(n0=n_0, n1=n_1)
            )

        # degrees of freedom = |V|-1
        dof0 = self.layers[0].degrees_of_freedom(assumption='ngrams')

        # degrees of freedom based on network assumption
        dof1 = self.layers[1].degrees_of_freedom(assumption='paths')

        Log.add('Log-Likelihood (k=0) = ' + str(likelihood_0), Severity.INFO)
        Log.add('Degrees of freedom (k=0) = ' + str(dof0), Severity.INFO)

        Log.add('Log-Likelihood (k=1) = ' + str(likelihood_1), Severity.INFO)
        Log.add('Degrees of freedom (k=1) = ' + str(dof0 + dof1), Severity.INFO)

        if method == 'AIC':
            ic0 = 2 * dof0 - 2 * likelihood_0
            ic1 = 2 * (dof0 + dof1) - 2 * likelihood_1
        elif method == 'AICc':
            dof10 = dof0 + dof1
            assert n_1 > dof10 - 2, \
                'Error: number of samples too small for model complexity'
            dof10 = dof0 + dof1
            ic0 = 2 * dof0 - 2 * likelihood_0 + (2 * (dof0 + 1) * (dof0 + 2)) / (n_0 - dof0 - 2)
            ic1 = 2 * dof10 - 2 * likelihood_1 + (2 * (dof10+1) * (dof10 + 2)) / (n_1 - dof10 - 2)
        elif method == 'BIC':
            ic0 = np.log(n_0) * dof0 - 2 * likelihood_0
            ic1 = np.log(n_1) * (dof0 + dof1) - 2 * likelihood_1
        else:
            raise PathpyError("Method check has not filtered out illegal "
                                  "method %s " % method)

        # if the AIC/AICc/BIC of the zero-order model is larger than that of the
        # first-order model, we do not reject the network hypothesis
        return ic0 > ic1, ic0, ic1



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : markov.py -- Markov processes
# Author    : Juergen Hackl <hackl@ibi.baug.ethz.ch>
# Creation  : 2018-07-20
# Time-stamp: <Fre 2018-07-20 10:29 juergen>
#
# Copyright (c) 2018 Juergen Hackl <hackl@ibi.baug.ethz.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================
import numpy as np
from cnet import logger
from cnet.utils.exceptions import CnetError, CnetNotImplemented
import cvxpy as cvx
log = logger(__name__)


def cost_matrix(n, mode='linear'):
    """Cost matrix indicating the costs for changing paths.

    The cost matrix C = (c_ij) is an indicator for some kind of degree of
    difficulty to transfer from path i to path j. This can be interpreted as the
    drivers utility to change from its preferable path.

    Parameters
    ----------
    n : int
        Number of paths which can be chosen.

    mode : str, optional (default = 'linear') Type of the cost function. Per
        default the costs increase linear, i.e. from node 1 to 2 the cost is 1
        and for 1 to 3 is 2. With mode = 'quadratic' the costs increase
        quadratic i.e. from 1 to 3 the costs are 2**2 = 4.

    Returns
    -------
    C : numpy.array
        A cost matrix indicating the costs of changing paths.

    TODO
    ----
    Add cost function dependent on the path travel costs.

    """
    # set up empty cost matrix
    C = np.zeros((n, n))

    # iterate through the elements of the matrix and assign cost values
    for i in range(0, n-1):
        for j in range(i+1, n):
            # TODO: replace j-i with the actual cost function.
            if mode == 'linear':
                C[i, j] = (j-i)
            else:
                C[i, j] = (j-i)**2
    # make the cost matrix symmetric
    C += C.T
    # return the cost matrix
    return C


def estimate_transition_matrix(x_1, x_2, C, verbose=False):
    """Estimation of the transition matrix given two states and a cost matrix.

    The estimation of the transition matrix expressed as a minimization
    problem, minimizing the costs for the user changing the paths.

    .. math::
       \min \sum_i \sum_j x_ij \cdot c_ij

    subjected to the constrains

    .. math::
       \sum_j x_ij &= x_1\\
       \sum_i x_ij &= x_2\\
       x_ij &\leg 0

    Matrix X is normalized to a row sum = 1

    .. math::
       x_ij / x_{1,i}

    Parameters
    ----------
    x_1 : list or numpy.array
        A vector with the initial state (e.g. at time t).

    x_2 : list or numpy.array
        A vector with the final state (e.g. at time t+1)

    C : numpy.array
        A cost matrix indicating the costs of changing paths.

    verbose : Boole, optional (default = False)
        If this option is enabled, information of the optimization problem is
        printed to the terminal. Per default this option is disabled.

    Returns
    -------
    T : numpy.array (or spare a matrix?)
        A Transition matrix describing the transition from x_1 to x_2.

    """

    # initialize vectors
    if isinstance(x_1, list):
        x_1 = np.asarray(x_1)
    elif not isinstance(x_1, np.ndarray):
        log.error('x_1 is not a vector in the proper format. '
                  'Please use a list or a numpy.array!')
        raise CnetError

    if isinstance(x_2, list):
        x_2 = np.asarray(x_2)
    elif not isinstance(x_2, np.ndarray):
        log.error('x_2 is not a vector in the proper format. '
                  'Please use a list or a numpy.array!')
        raise CnetError

    # number of nodes
    n = len(x_1)

    # check size of the other elements
    if len(x_2) != n:
        log.error('Size of the vectors does not match {} != {}.'
                  ''.format(len(x_2), n))
        raise CnetError
    if C.shape != (n, n):
        log.error('Size of the cost matrix does not match {} != ({}, {}).'
                  ''.format(C.shape, n, n))
        raise CnetError

    # Normalize vector x_1 to a row sum of 1
    # TODO: figure out if normalization is an advantage or not?
    # x_1 = x_1 / np.sum(x_1)

    # Scale vector to the same row sum as x_1
    # Otherwise no optimal solution can be found
    if np.sum(x_1) != np.sum(x_2):
        x_2 = x_2 / np.sum(x_2) * np.sum(x_1)

    # Transition matrix to estimate
    X = cvx.Variable((n, n))

    # Objective function
    objective = cvx.Minimize(cvx.sum(cvx.multiply(C, X)))

    # Constraints
    constraints = [X >= 0,
                   cvx.sum(X, axis=1) == x_1,
                   cvx.sum(X, axis=0) == x_2]

    # Minimization problem
    prob = cvx.Problem(objective, constraints)

    # Result form the optimization process
    res = prob.solve(verbose=verbose)

    # create temporary result matrix
    _X = np.round(X.value, 0)

    # log.debug(np.sum(_X, axis=1) == x_1)
    # log.debug(np.sum(_X, axis=0) == x_2)

    # Convert estimated matrix X to probabilities
    T = np.divide(_X, x_1[:, None], out=np.zeros_like(_X),
                  where=x_1[:, None] != 0)
    # TODO: add additional checks for the transition matrix.
    # TODO: try out to return a sparse matrix
    return T

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



# -*- coding: utf-8 -*-

#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net
"""
This class can be used to calculate path statistics based on
origin-destination data available for a known network topology.
The path statistics generated from such data will be based on
the assumption that each observed path from an origin to a destination
node follows a shortest path in the network topology.
"""
from collections import defaultdict
import operator

import numpy as _np
import scipy.sparse as sparse
import scipy.sparse.linalg as sla
import scipy.linalg as la
import scipy as sp

from pathpy .utils import Log, Severity
from pathpy import HigherOrderNetwork
from pathpy.algorithms.shortest_paths import *

from pathpy.utils import PathpyNotImplemented


__all__ = ['eigenvalue_gap', 'fiedler_vector_sparse', 'fiedler_vector_dense', 'algebraic_connectivity']

def eigenvalue_gap(network, include_sub_paths=True, lanczos_vectors=15, maxiter=20):
    """Returns the eigenvalue gap of the transition matrix.

    Parameters
    ----------
    network
    include_sub_paths: bool
        whether or not to include subpath statistics in the calculation of transition
        probabilities.
    lanczos_vectors: int
        number of Lanczos vectors to be used in the approximate
        calculation of eigenvectors and eigenvalues. This maps to the ncv parameter
        of scipy's underlying function eigs.
    maxiter: int
        scaling factor for the number of iterations to be used in the
        approximate calculation of eigenvectors and eigenvalues. The number of iterations
        passed to scipy's underlying eigs function will be n*maxiter where n is the
        number of rows/columns of the Laplacian matrix.

    Returns
    -------
    float
    """
    assert isinstance(network, HigherOrderNetwork), \
        "network must be an instance of HigherOrderNetwork"
    # NOTE to myself: most of the time goes for construction of the 2nd order
    # NOTE            null graph, then for the 2nd order null transition matrix

    Log.add('Calculating eigenvalue gap ... ', Severity.INFO)

    # Build transition matrices
    trans_mat = network.transition_matrix(include_sub_paths)

    # Compute the two largest eigenvalues
    # NOTE: ncv sets additional auxiliary eigenvectors that are computed
    # NOTE: in order to be more confident to actually find the one with the largest
    # NOTE: magnitude, see https://github.com/scipy/scipy/issues/4987
    eig_vals = sla.eigs(trans_mat, which="LM", k=2, ncv=lanczos_vectors,
                        return_eigenvectors=False, maxiter=maxiter)
    eigen_values2_sorted = _np.sort(-_np.absolute(eig_vals))

    Log.add('finished.', Severity.INFO)

    return _np.abs(eigen_values2_sorted[1])


def fiedler_vector_sparse(network, normalized=True, lanczos_vectors=15, maxiter=20):
    """Returns the (sparse) Fiedler vector of the higher-order network. The Fiedler
    vector can be used for a spectral bisection of the network.

    Note that sparse linear algebra for eigenvalue problems with small eigenvalues
    is problematic in terms of numerical stability. Consider using the dense version
    of this method in this case. Note also that the sparse Fiedler vector might be scaled
    by a factor (-1) compared to the dense version.

    Parameters
    ----------
    network
    normalized: bool
        whether (default) or not to normalize the fiedler vector.
        Normalization is done such that the sum of squares equals one in order to
        get reasonable values as entries might be positive and negative.
    lanczos_vectors: int
        number of Lanczos vectors to be used in the approximate
        calculation of eigenvectors and eigenvalues. This maps to the ncv parameter
        of scipy's underlying function eigs.
    maxiter: int
        scaling factor for the number of iterations to be used in the
        approximate calculation of eigenvectors and eigenvalues. The number of iterations
        passed to scipy's underlying eigs function will be n*maxiter where n is the
        number of rows/columns of the Laplacian matrix.

    Returns
    -------

    """
    assert isinstance(network, HigherOrderNetwork), \
        "network must be an instance of HigherOrderNetwork"
    # NOTE: The transposed matrix is needed to get the "left" eigenvectors
    lapl_mat = network.laplacian_matrix()

    # NOTE: ncv sets additional auxiliary eigenvectors that are computed
    # NOTE: in order to be more confident to find the one with the largest
    # NOTE: magnitude, see https://github.com/scipy/scipy/issues/4987
    maxiter = maxiter * lapl_mat.get_shape()[0]
    w = sla.eigs(lapl_mat, k=2, which="SM", ncv=lanczos_vectors,
                 return_eigenvectors=False, maxiter=maxiter)

    # compute a sparse LU decomposition and solve for the eigenvector
    # corresponding to the second largest eigenvalue
    lapl_n = lapl_mat.get_shape()[0]
    fiedler_v = _np.ones(lapl_n)
    eigen_value = _np.sort(_np.abs(w))[1]
    mat = (lapl_mat[1:lapl_n, :].tocsc()[:, 1:lapl_n] -
           sparse.identity(lapl_n - 1).multiply(eigen_value))
    fiedler_v[1:lapl_n] = mat[0, :].toarray()

    lu_decom = sla.splu(mat)
    fiedler_v[1:lapl_n] = lu_decom.solve(fiedler_v[1:lapl_n])

    if normalized:
        fiedler_v /= _np.sqrt(_np.inner(fiedler_v, fiedler_v))
    return fiedler_v


def fiedler_vector_dense(network):
    """Returns the (dense)Fiedler vector of the higher-order network.
    The Fiedler vector can be used for a spectral bisection of the network.

    Parameters
    ----------
    network: HigherOrderNetwork

    Returns
    -------

    """
    assert isinstance(network, HigherOrderNetwork), \
        "network must be an instance of HigherOrderNetwork"
    # NOTE: The Laplacian is transposed for the sparse case to get the left
    # NOTE: eigenvalue.
    lapl_mat = network.laplacian_matrix()
    # convert to dense matrix and transpose again to have the un-transposed
    # laplacian again.
    laplacian_transposed = lapl_mat.todense().transpose()
    w, v = la.eig(laplacian_transposed, right=False, left=True)

    return v[:, _np.argsort(_np.absolute(w))][:, 1]


def algebraic_connectivity(network, lanczos_vectors=15, maxiter=20):
    """

    Parameters
    ----------
    network: HigherOrderNetwork
    lanczos_vectors: int
        number of Lanczos vectors to be used in the approximate calculation of
        eigenvectors and eigenvalues. This maps to the ncv parameter of scipy's underlying
        function eigs.
    maxiter: int
        scaling factor for the number of iterations to be used in the approximate
        calculation of eigenvectors and eigenvalues. The number of iterations passed to
        scipy's underlying eigs function will be n*maxiter where n is the number of
        rows/columns of the Laplacian matrix.

    Returns
    -------

    """
    assert isinstance(network, HigherOrderNetwork), \
        "network must be an instance of HigherOrderNetwork"
    Log.add('Calculating algebraic connectivity ... ', Severity.INFO)

    lapl_mat = network.laplacian_matrix()
    # NOTE: ncv sets additional auxiliary eigenvectors that are computed
    # NOTE: in order to be more confident to find the one with the largest
    # NOTE: magnitude, see https://github.com/scipy/scipy/issues/4987
    w = sla.eigs(lapl_mat, which="SM", k=2, ncv=lanczos_vectors,
                 return_eigenvectors=False, maxiter=maxiter)
    eigen_values_sorted = _np.sort(_np.absolute(w))

    Log.add('finished.', Severity.INFO)

    # TODO: result is unstable, it looks like it depends on a "warm start"
    # (i.e. run after other eigen velue calculations) see test_algebraic_connectivity
    # problems with order k=3

    return _np.abs(eigen_values_sorted[1])



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net
from collections import defaultdict
import sys
import copy

import numpy as np
from pathpy.utils import Log, Severity
from pathpy.utils.exceptions import PathpyError
from pathpy.utils.default_containers import nested_zero_default as _nested_zero_default
from pathpy.utils.default_containers import zero_array_default as _zero_array_default


class Paths:
    """
    Instances of this class represent path statistics
    which can be analyzed using higher- and multi-order network
    models. The origin of the path statistics can be (i) n-gram
    files which provide us with a list of paths in terms of n-grams
    of varying lengths, or (ii) a temporal network instance which
    provides us with a set of time-respecting paths based on a
    given maximum time difference delta.
    """

    def __init__(self, separator=','):
        """
        Creates an empty Paths object
        """

        # A dictionary of paths that has the following structure:
        # - paths[k] is a dictionary containing all paths of length k,
        #    indexed by a path tuple p = (u,v,w,...)
        # - for each tuple p of length k, paths[k][p] contains a tuple
        #    (i,j) where i refers to the number of times p occurs as a
        #    subpath of a longer path, and j refers to the number of times p
        #    occurs as a *real* or *longest* path (i.e. not being a subpath
        #    of a longer path)
        self.paths = _nested_zero_default()

        # The character used to separate nodes on paths
        self.separator = separator

        # This can be used to limit the calculation of sub path statistics to a given
        # maximum length. This is useful, as the statistics of sub paths of length k
        # are only needed to fit a higher-order model with order k. Hence, if we know
        # that the model selection is limited to a given maximum order K, we can safely
        #  set the maximum sub path length to K. By default, sub paths of any length
        # will be calculated. Note that, independent of the sub path calculation
        # longest path of any length will be considered in the likelihood calculation!
        self.max_subpath_length = sys.maxsize

    def summary(self):
        """

        Returns
        -------
        str
            Returns a string containing basic summary info of this Paths instance
        """
        total_paths = []
        sub_path_sum = []
        l_path_sum = []
        max_path_length = 0
        average_length = 0
        for k in sorted(self.paths):
            paths_ = self.paths[k]
            values_ = np.array(list(paths_.values()))
            v_0 = np.sum(values_[:, 0])
            v_1 = np.sum(values_[:, 1])
            total_paths += [v_0 + v_1]
            sub_path_sum += [v_0]
            l_path_sum += [v_1]
            average_length += v_1 * k
            if paths_:
                max_path_length = max(max_path_length, k)
        if np.sum(l_path_sum) > 0:
            average_length = average_length / np.sum(l_path_sum)

        summary_fmt = (
            "Total path count: \t\t{lpsum} \n"
            "[Unique / Sub paths / Total]: \t[{unique_paths} / {spsum} / {total_paths}]\n"
            "Nodes:\t\t\t\t{len_nodes} \n"
            "Edges:\t\t\t\t{len_edges}\n"
            "Max. path length:\t\t{maxL}\n"
            "Avg path length:\t\t{avgL} \n"
        )

        k_path_info_fmt = 'Paths of length k = {k}\t\t{lpsum} ' \
                          '[ {unique_paths_longer} / {spsum} / {total_paths} ]\n'

        # Count number of nodes and edges
        if 0 not in self.paths:
            len_0 = 0
        else:
            len_0 = len(self.paths[0])
        if 1 not in self.paths:
            len_1 = 0
        else:
            len_1 = len(self.paths[1])

        summary_info = {
            "lpsum": np.sum(l_path_sum),
            "unique_paths": self.unique_paths(),
            "spsum": np.sum(sub_path_sum),
            "total_paths": np.sum(total_paths),
            "len_nodes": len_0,
            "len_edges": len_1,
            "maxL": max_path_length,
            "avgL": average_length
        }

        summary = summary_fmt.format(**summary_info)

        for k in sorted(self.paths):
            k_info = k_path_info_fmt.format(
                k=k, lpsum=l_path_sum[k], spsum=sub_path_sum[k],
                total_paths=total_paths[k],
                unique_paths_longer=self.unique_paths(l=k, consider_longer_paths=False)
            )
            summary += k_info

        return summary

    def path_lengths(self):
        """compute the length of all paths

        Returns
        -------
        dict
            Returns a dictionary containing the distribution of path lengths
            in this Path object. In the returned dictionary, entry
            lengths ``k`` is a ``numpy.array`` ``x`` where
            ``x[0]`` is the number of sub paths with length ``k``, and ``x[1]``
            is the number of (longest) paths with length ``k``


        """
        lengths = _zero_array_default()

        for k in self.paths:
            for p in self.paths[k]:
                lengths[k] += self.paths[k][p]
        return lengths

    def __add__(self, other):
        """add path statistics of one object to the other

        Parameters
        ----------
        other : Paths

        Returns
        -------
        Paths
            Default operator +, which returns the sum of two Path objects
        """
        p_sum = Paths()
        p_sum.paths = copy.deepcopy(self.paths)
        for p_length in other.paths:
            for p in other.paths[p_length]:
                p_sum.paths[p_length][p] += other.paths[p_length][p]
        return p_sum

    def __iadd__(self, other):
        """in place addition avoids unnecessary copies of the object

        Parameters
        ----------
        other

        Returns
        -------
        None

        """
        for p_length in other.paths:
            for p in other.paths[p_length]:
                self.paths[p_length][p] += other.paths[p_length][p]
        return self

    def __mul__(self, factor):
        """multiplies all path statistics by factor

        Parameters
        ----------
        factor

        Returns
        -------
        a Paths object with multiplied frequencies

        """
        p_mult = Paths()
        for p_length in self.paths:
            for p in self.paths[p_length]:
                p_mult.paths[p_length][p] = self.paths[p_length][p] * factor

        return p_mult

    def __rmul__(self, factor):
        """right multiply"""
        return self * factor

    def __imul__(self, factor):
        """in-place scaling of path statistics

        Parameters
        ----------
        factor

        Returns
        -------
        None


        """
        for l in self.paths:
            for p in self.paths[l]:
                self.paths[l][p] = self.paths[l][p] * factor

        return self

    def sequence(self, stop_char='|'):
        """

        Parameters
        ----------
        stop_char : str
            the character used to separate paths

        Returns
        -------
        tuple:
            Returns a single sequence in which all paths have been concatenated.
            Individual paths are separated by a stop character.
        """
        Log.add('Concatenating paths to sequence ...')
        sequence = []
        for p_length in self.paths:
            for p in self.paths[p_length]:
                segment = []
                for s in p:
                    segment.append(s)
                if stop_char != '':
                    segment.append(stop_char)
                for _ in range(int(self.paths[p_length][p][1])):
                    sequence += segment

        Log.add('finished')
        return sequence

    def unique_paths(self, l=0, consider_longer_paths=True):
        """Returns the number of unique paths of a given length l (and possibly longer)

        Parameters
        ----------
        l : int
            the (inclusive) maximum length up to which path shall be counted.
        consider_longer_paths : bool
            TODO: add parameter description

        Returns
        -------
        int
            number of unique paths satisfying parameter ``l``
        """
        num_l = 0.0
        if not self.paths:
            return num_l

        max_length = l
        if consider_longer_paths:
            max_length = max(self.paths) if self.paths else 0
        for j in range(l, max_length + 1):
            for p in self.paths[j]:
                if self.paths[j][p][1] > 0:
                    num_l += 1.0
        return num_l

    def __str__(self):
        """
        Returns the default string representation of
        this Paths instance
        """
        return self.summary()

    @property
    def nodes(self):
        """
        Returns the list of nodes for the underlying
        set of paths
        """
        nodes = set()
        for p in self.paths[0]:
            nodes.add(p[0])
        return nodes

    @staticmethod
    def read_edges(filename, separator=',', weight=False, undirected=False,
                   maxlines=None, expand_sub_paths=True, max_subpath_length=None):
        """
        Read path in edgelist format

        Reads data from a file containing multiple lines of *edges* of the
        form "v,w,frequency,X" (where frequency is optional and X are
        arbitrary additional columns). The default separating character ','
        can be changed. In order to calculate the statistics of paths of any length,
        by default all subpaths of length 0 (i.e. single nodes) contained in an edge
        will be considered.

        Parameters
        ----------
        filename : str
            path to edgelist file
        separator : str
            character separating the nodes
        weight : bool
            is a weight given? if ``True`` it is the last element in the edge
            (i.e. ``a,b,2``)
        undirected : bool
            are the edges directed or undirected
        maxlines : int
            number of lines to read (useful to test large files)
        expand_sub_paths : bool
        max_subpath_length : int (default None)
            maximum length for subpaths to consider, ``None`` means the entire file is
            read
            TODO: this parameter is unused.

        Returns
        -------
        Paths
            a ``Paths`` object obtained from the edgelist
        """
        p = Paths()

        p.separator = separator
        p.max_subpath_length = sys.maxsize

        with open(filename, 'r') as f:
            Log.add('Reading edge data ... ')
            for n, line in enumerate(f):
                fields = line.rstrip().split(separator)
                assert len(fields) >= 2, 'Error: malformed line: {0}'.format(line)
                path = (fields[0], fields[1])

                frequency = int(fields[2]) if weight else 1

                p.paths[1][path] += (0, frequency)
                if undirected:
                    reverse_path = (fields[1], fields[0])
                    p.paths[1][reverse_path] += (0, frequency)

                if maxlines is not None and n >= maxlines:
                    break
        if expand_sub_paths:
            p.expand_subpaths()
        Log.add('finished.')

        return p

    @classmethod
    def read_file(cls, filename, separator=',', frequency=False, maxlines=sys.maxsize,
                  max_ngram_length=sys.maxsize, expand_sub_paths=True,
                  max_subpath_length=sys.maxsize):
        """Read path data in ngram format.

        Reads path data from a file containing multiple lines of n-grams of the form
        ``a,b,c,d,frequency`` (where frequency is optional).
        The default separating character ',' can be changed. Each n-gram will be
        interpreted as a path of
        length n-1, i.e. bigrams a,b are considered as path of length one, trigrams a,
        b,c as path of length two, etc. In order to calculate the statistics of paths
        of any length, by default all subpaths of length k < n-1 contained in an n-gram
        will be considered. I.e. for n=4 the four-gram a,b,c,d will be considered as a
        single (longest) path of length n-1 = 3 and three subpaths a->b, b->c, c->d of
        length k=1 and two subpaths a->b->c amd b->c->d of length k=2 will be
        additionally counted.

        Parameters
        ----------
        filename : str
            path to the n-gram file to read the data from
        separator : str
            the character used to separate nodes on the path, i.e. using a
            separator character of ';' n-grams are represented as ``a;b;c;...``
        frequency : bool
            if set to ``True``, the last entry in each n-gram will be interpreted as
            weight (i.e. frequency of the path), e.g. ``a,b,c,d,4`` means that four-gram
            ``a,b,c,d`` has weight four. ``False`` by default, which means each path
            occurrence is assigned a default weight of 1 (adding weights for multiple
            occurrences).
        maxlines : int
            number of lines/n-grams to read, if left at None the whole file is read in.
        max_ngram_length : int
            The maximum n for the n-grams to read, i.e. setting max_ngram_length to 15
            will ignore
            all n-grams of length 16 and longer, which means that only paths up to length
            n-1 are considered.
        expand_sub_paths : bool
            by default all subpaths of the given n-grams are generated, i.e.
            for an input file with a single trigram a;b;c a path a->b->c of length two
            will be generated as well as two subpaths a->b and b->c of length one
        max_subpath_length : int

        Returns
        -------
        Paths
            a ``Paths`` object obtained from the n-grams file
        """
        assert filename != "", 'Empty filename given'

        # If subpath expansion is applied, we keep the information how many times a path
        # has been observed as a subpath, and how many times as a "real" path

        p = cls()

        p.max_subpath_length = max_subpath_length
        p.separator = separator
        max_length = 0

        with open(filename, 'r') as f:
            Log.add('Reading ngram data ... ')
            line = f.readline()
            n = 1
            while line and n <= maxlines:
                fields = line.rstrip().split(separator)
                path = ()
                # Add frequency of "real" path to second component of occurrence counter
                if frequency:
                    for i in range(0, len(fields) - 1):
                        # Omit empty fields
                        v = fields[i].strip()
                        if v:
                            path += (v,)                                         
                    freq = float(fields[len(fields) - 1])
                    if freq >0:
                        if len(path) <= max_ngram_length:
                            p.paths[len(path) - 1][path] += (0, freq)
                            max_length = max(max_length, len(path) - 1)
                        else:  # cut path at max_ngram_length
                            mnl = max_ngram_length
                            p.paths[mnl - 1][path[:mnl]] += (0, freq)
                            max_length = max(max_length, max_ngram_length - 1)
                    else:
                        Log.add('Non-positive path count in line {0}'.format(n), Severity.WARNING)
                else:
                    for field in fields:
                        # Omit empty fields
                        v = field.strip()
                        if v:
                            path += (v,)
                    if len(path) <= max_ngram_length:
                        p.paths[len(path) - 1][path] += (0, 1)
                        max_length = max(max_length, len(path) - 1)
                    else:  # cut path at max_ngram_length
                        p.paths[max_ngram_length - 1][path[:max_ngram_length]] += (0, 1)
                        max_length = max(max_length, max_ngram_length - 1)
                line = f.readline()
                n += 1
        # end of with open()
        Log.add(
            'finished. Read ' + str(n - 1) + ' paths with maximum length ' + str(max_length))

        if expand_sub_paths:
            p.expand_subpaths()
        Log.add('finished.')

        return p

    def write_file(self, filename, separator=','):
        """Writes path statistics data to a file. Each line in this file captures a
        longest path (v0,v1,...,vl), as well as its frequency f as follows

        Parameters
        ----------
        filename: str
            name of the file to write to
        separator: str
            character that shall be used to separate nodes and frequencies

        Returns
        -------

        """
        with open(filename, 'w') as f:
            for p_length in self.paths:
                for p in self.paths[p_length]:
                    if self.paths[p_length][p][1] > 0:
                        line = ""
                        for x in p:
                            line += x
                            line += separator
                        line += str(self.paths[p_length][p][1])
                        f.write(line + '\n')
        f.close()

    @property
    def observation_count(self):
        """
        Returns the total number of observed pathways of any length
        (includes multiple observations for paths with a frequency weight)
        """

        obs_count = 0
        for k in self.paths:
            for p in self.paths[k]:
                obs_count += self.paths[k][p][1]
        return obs_count

    def expand_subpaths(self):
        """
        This function implements the sub path expansion, i.e.
        for a four-gram a,b,c,d, the paths a->b, b->c, c->d of
        length one and the paths a->b->c and b->c->d of length
        two will be counted.

        This process will consider restrictions to the maximum
        sub path length defined in self.max_subpath_length
        """

        # nothing to see here ...
        if not self.paths:
            return

        Log.add('Calculating sub path statistics ... ')

        # the expansion of all subpaths in paths with a maximum path length of maxL
        # necessarily generates paths of *any* length up to MaxL.
        # Forcing the generation of all these indices here, prevents us
        # from mutating indices during subpath creation. The fact that indices are
        # immutable allows us to use efficient iterators and prevent unnecessarily copying

        # Thanks to the use of defaultdict, the following trick will prevent us from
        # repeatedly testing whether l already exists as a key
        for p_length in range(max(self.paths)):
            self.paths[p_length] = self.paths[p_length]

        # expand subpaths in paths of any length ...
        for path_length in self.paths:
            for path, value in self.paths[path_length].items():

                # The frequency is given by the number of occurrences as longest
                # path, which is stored in the second entry of the numpy array
                frequency = value[1]

                # compute maximum length of sub paths to consider
                # (maximum up to pathLength)
                max_length = min(self.max_subpath_length + 1, path_length)

                # Generate all subpaths of length k for k = 0 to k = max_len-1 (inclusive)
                for k in range(max_length):
                    # Generate subpaths of length k for all start indices s
                    # for s = 0 to s = pathLength-k (inclusive)
                    for s in range(path_length - k + 1):
                        # Add frequency as a subpath to *first* entry of occurrence
                        # counter
                        path_slice = path[s:s + k + 1]
                        self.paths[k][path_slice][0] += frequency

    def add_path_tuple(self, path, expand_subpaths=True, frequency=(0, 1)):
        """Adds a tuple of elements as a path. If the elements are not strings,
        a conversion to strings will be made. This function can be used to
        to set custom subpath statistics, via the frequency tuple (see below).

        Parameters
        ----------
        path: tuple
            The path tuple to be added, e.g. ('a', 'b', 'c')
        expand_subpaths: bool
            Whether or not to calculate subpath statistics for this path
        frequency: tuple
            A tuple (x,y) indicating the frequency of this path as subpath
            (first component) and longest path (second component). Default is (0,1).

        Returns
        -------

        """
        assert path, 'Error: paths needs to contain at least one element'

        for x in path:
            if isinstance(x, str) and self.separator in x:
                raise PathpyError('Error: Node name contains separator character. '
                                  'Choose different separator.')

        path_str = path if isinstance(path, str) else tuple(map(str, path))

        self.paths[len(path) - 1][path_str] += frequency

        if expand_subpaths:

            max_length = min(self.max_subpath_length + 1, len(path_str) - 1)

            for k in range(0, max_length):
                for s in range(len(path_str) - k):
                    # for all start indices from 0 to n-k

                    subpath = ()
                    # construct subpath
                    for i in range(s, s + k + 1):
                        subpath += (path_str[i],)
                    # add subpath weight to first component of occurrences
                    self.paths[k][subpath][0] += frequency[1]

    def add_path_ngram(self, ngram, frequency=(0, 1), separator=',', expand_subpaths=True):
        """Adds the path(s) of a single n-gram to the path statistics object.

        Parameters
        ----------
        ngram: str
            An ngram representing a path between nodes, separated by the separator
            character, e.g. the 4-gram a;b;c;d represents a path of length three
            (with separator ';')
        frequency: tuple, int
            the number of occurrences (i.e. frequency) of the ngram (n_subpaths, n_observed)
            the default is (0, 1) (i.e. 0 subpaths and one observed longest path). If an integer x
            is passed, it will be automatically converted to (0, x).
        separator: str
            The character used as separator for the ngrams (';' by default)
        expand_subpaths: bool
            by default all subpaths of the given ngram are generated, i.e.
            for the trigram a;b;c a path a->b->c of length two will be generated
            as well as two subpaths a->b and b->c of length one        

        """
        path = tuple(ngram.split(separator))
        for x in path:
            if isinstance(x, str) and self.separator in x:
                raise PathpyError('Error: Node name contains separator character.'
                                  'Choose different separator.')

        path_length = len(path) - 1

        # add the occurrences as *longest* path to the second component of the numpy array
        if isinstance(frequency, int):
            frequency = (0, frequency)
        self.paths[path_length][path][1] += frequency[1]

        if expand_subpaths:
            max_length = min(self.max_subpath_length + 1, path_length)
            for k in range(0, max_length):
                for s in range(len(path) - k):
                    # for all start indices from 0 to n-k

                    subpath = ()
                    # construct subpath
                    for i in range(s, s + k + 1):
                        subpath += (path[i],)
                    # add subpath weight to first component of occurrences
                    self.paths[k][subpath][0] += frequency[1]

    @staticmethod
    def contained_paths(p, node_filter):
        """Returns the set of maximum-length sub-paths of the path p, which only contain
        nodes that appear in the node_filter. As an example, for the path (a,b,c,d,e,f,g)
        and a node_filter [a,b,d,f,g], the method will return [(a,b), (d,), (f,g)].

        Parameters
        ----------
        p: tuple
            a path tuple to check for contained paths
        node_filter: set
            a set of nodes to which the contained paths should be limited

        Returns
        -------
        """

        contained_paths = []
        current_path = ()
        for node in p:
            if node in node_filter:
                current_path += (node,)
            else:
                if current_path:
                    contained_paths.append(current_path)
                    current_path = ()
        if current_path:
            contained_paths.append(current_path)

        return contained_paths

    def filter_nodes(self, node_filter, min_length=0, max_length=sys.maxsize):
        """Returns a new paths object which contains only paths between nodes in a given
        filter set. For each of the paths in the current Paths object, the set of
        maximally contained subpaths between nodes in node_filter is extracted.
        This method is useful when studying (sub-)paths passing through a subset of nodes.

        Parameters
        ----------
        node_filter: set
            the nodes for which paths with be extracted from the current set of paths
        min_length: int
            the minimum length of paths to extract (default 0)
        max_length: int
            the maximum length of paths to extract (default sys.maxsize)

        Returns
        -------
        Paths
        """
        p = Paths()
        for p_length in self.paths:
            for x in self.paths[p_length]:
                if self.paths[p_length][x][1] > 0:
                    # determine all contained subpaths which only pass through
                    # nodes in node_filter
                    contained = Paths.contained_paths(x, node_filter)
                    for s in contained:
                        if min_length <= len(s) - 1 <= max_length:
                            freq = (0, self.paths[p_length][x][1])
                            p.add_path_tuple(s, expand_subpaths=True, frequency=freq)
        return p

    def project_paths(self, mapping):
        """Returns a new path object in which nodes have been mapped to different labels
        given by an arbitrary mapping function. For instance, for the mapping
        {'a': 'x', 'b': 'x', 'c': 'y', 'd': 'y'} the path (a,b,c,d) is mapped to
        (x,x,y,y). This is useful, e.g., to map page page click streams to topic
        click streams, using a mapping from pages to topics.

        Parameters
        ----------
        mapping: dict
            a dictionary that maps nodes to the new labels

        Returns
        -------

        """
        p = Paths()
        p.max_subpath_length = self.max_subpath_length
        for p_length in self.paths:
            for x in self.paths[p_length]:
                # if this path occurred as longest path
                if self.paths[p_length][x][1] > 0:
                    # construct projected path
                    new_p = ()
                    for v in x:
                        new_p += (mapping[v],)
                    # add to new path object and expand sub paths
                    freq = (0, self.paths[p_length][x][1])
                    p.add_path_tuple(new_p, expand_subpaths=True, frequency=freq)
        return p



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : exceptions.py 
# Creation  : 30 Apr 2018
# Time-stamp: <Fre 2018-05-04 10:14 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$ 
#
# Description : Module with the base exceptions
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 
# =============================================================================


class CnetException(Exception):
    """Base class for all cnet specific exceptions."""

class CnetError(CnetException):
    """Exception for a serious error in cnet"""

class CnetNotImplemented(CnetException):
    """Exception for procedure not implemented in cnet."""


    
# =============================================================================
# eof
#
# Local Variables: 
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:  



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

import collections as _co

import numpy as _np

import scipy.sparse as _sparse
import scipy.sparse.linalg as _sla

from pathpy.utils import Log, Severity
from pathpy.utils.exceptions import PathpyError, PathpyNotImplemented


class Network:
    """
    Instances of this class capture a graph or network
    that can be directed, undirected, unweighted or weighted
    """

    def __init__(self, directed=False):
        """
        Generates an empty network.
        """

        # Boolean value that inidcates whether the network is directed or undirected
        self.directed = directed

        # A dictionary containing nodes as well as node properties
        self.nodes = _co.defaultdict(dict)

        # A dictionary containing edges as well as edge properties
        if not directed:
            self.edges = UnorderedDict()
        else:
            self.edges = _co.defaultdict(dict)

        # A dictionary containing the sets of successors of all nodes
        self.successors = _co.defaultdict(set)

        # A dictionary containing the sets of predecessors of all nodes
        self.predecessors = _co.defaultdict(set)


    @classmethod
    def read_edges(cls, filename, separator=',', weighted=False, directed=False, header=False):
        """
        Reads a network from an edge list file

        Reads data from a file containing multiple lines of *edges* of the
        form "v,w,frequency,X" (where frequency is optional and X are
        arbitrary additional columns). The default separating character ','
        can be changed. In order to calculate the statistics of paths of any length,
        by default all subpaths of length 0 (i.e. single nodes) contained in an edge
        will be considered.

        Parameters
        ----------
        filename : str
            path to edgelist file
        separator : str
            character separating the nodes
        weighted : bool
            is a weight given? if ``True`` it is the last element in the edge
            (i.e. ``a,b,2``)
        directed : bool
            are the edges directed or undirected
        header: bool
            if true skip the first row, useful if header row in file

        Returns
        -------
        Network
            a ``Network`` object obtained from the edgelist
        """
        net = cls(directed)

        with open(filename, 'r') as f:
            Log.add('Reading edge list ... ')
            header_offset = 0
            if header:
                f.readline()
                header_offset = 1

            for n, line in enumerate(f):
                fields = line.rstrip().split(separator)
                fields = [field.strip() for field in fields]
                if len(fields) < 2:
                    Log.add('Ignoring malformed line {0}: {1}'.format(n, line+header_offset), Severity.WARNING)
                else:
                    if weighted:
                        net.add_edge(fields[0], fields[1], weight=int(fields[2]))
                    else:
                        net.add_edge(fields[0], fields[1])

        Log.add('finished.')

        return net


    @classmethod
    def from_sqlite(cls, cursor, directed=True):
        """Reads links from an SQLite cursor and returns a new instance of
        the class Network. The cursor is assumed to refer to a table that
        minimally has two columns

                source target

        and where each row refers to a link. Any additional columns will be used as
        edge properties

        Important: Since columns are accessed by name this function requires that a
        row factory object is set for the SQLite connection prior to cursor creation,
        i.e. you should set

                connection.row_factory = sqlite3.Row

        Parameters
        ----------
        cursor:
            The SQLite cursor to fetch rows
        directed: bool

        Returns
        -------

        """
        n = cls(directed=directed)

        assert cursor.connection.row_factory, \
            'Cannot access columns by name. Please set ' \
            'connection.row_factory = sqlite3.Row before creating DB cursor.'

        Log.add('Retrieving links from database ...')

        for row in cursor:
            n.add_edge(str(row['source']), str(row['target']))

        return n


    @classmethod
    def from_paths(cls, paths):
        
        network = cls(directed=True)

        # check all sub-paths of length one
        for p in paths.paths[1]:
            network.add_edge(p[0], p[1], weight=paths.paths[1][p].sum())

        return network


    def to_unweighted(self):
        """
        Returns an unweighted copy of a directed or undirected network.
        In this copy all edge and node properties of the original network 
        are removed, but the directionality of links is retained.
        """
        n = Network(directed = self.directed)

        for (v,w) in self.edges:
            n.add_edge(v, w)
        return n


    def to_undirected(self):
        """
        Returns an undirected copy of the network, in which all 
        node and edge properties are removed.
        """
        n = Network(directed = False)

        for (v,w) in self.edges:
            if v!=w:
                n.add_edge(v, w)
        return n


    def add_node(self, v, **kwargs):
        """
        Adds a node to a network
        """
        if v not in self.nodes:
            self.nodes[v] = {**self.nodes[v], **kwargs}

            # set default values if not set already
            if 'inweight' not in self.nodes[v]:
                self.nodes[v]['inweight'] = 0.0
            if 'outweight' not in self.nodes[v]:
                self.nodes[v]['outweight'] = 0.0
            if self.directed:
                self.nodes[v]['indegree'] = 0
                self.nodes[v]['outdegree'] = 0
            else:
                self.nodes[v]['degree'] = 0

    def remove_node(self, v):
        """
        Removes a node from the network
        """
        if v in self.nodes:
            # remove all incident edges and update neighbors
            if not self.directed:
                for w in self.successors[v]:
                    edge = (v, w)
                    self.nodes[w]['degree'] -= 1
                    self.nodes[w]['inweight'] -= self.edges[edge]['weight']
                    self.nodes[w]['outweight'] -= self.edges[edge]['weight']
                    self.successors[w].remove(v)
                    self.predecessors[w].remove(v)
                    del self.edges[edge]
            else:
                for w in self.successors[v]:
                    self.nodes[w]['indegree'] -= 1
                    self.nodes[w]['inweight'] -= self.edges[(v, w)]['weight']
                    self.predecessors[w].remove(v)
                    del self.edges[(v, w)]
                for w in self.predecessors[v]:
                    self.nodes[w]['outdegree'] -= 1
                    self.nodes[w]['outweight'] -= self.edges[(w, v)]['weight']
                    self.successors[w].remove(v)
                    del self.edges[(w, v)]
            del self.nodes[v]
            del self.successors[v]
            del self.predecessors[v]


    def remove_edge(self, source, target):
        """
        remove an edge from the network

        Parameters
        ----------
        source
        target
        """
        if not (source in self.nodes and target in self.nodes):
            return None

        if self.directed:
            # take care of source
            self.nodes[source]['outdegree'] -= 1
            self.nodes[source]['outweight'] -= self.edges[(source, target)]['weight']
            self.successors[source].remove(target)

            # take care of target
            self.nodes[target]['indegree'] -= 1
            self.nodes[target]['inweight'] -= self.edges[(source, target)]['weight']
            self.predecessors[target].remove(source)

            del self.edges[(source, target)]
        else:
            # take care of source
            self.nodes[source]['degree'] -= 1
            self.nodes[source]['outweight'] -= self.edges[(source, target)]['weight']
            self.nodes[source]['inweight'] -= self.edges[(source, target)]['weight']
            self.successors[source].remove(target)
            self.predecessors[source].remove(target)

            # take care of target
            self.nodes[target]['degree'] -= 1
            self.nodes[target]['outweight'] -= self.edges[(source, target)]['weight']
            self.nodes[target]['inweight'] -= self.edges[(source, target)]['weight']
            self.successors[target].remove(source)
            self.predecessors[target].remove(source)

            del self.edges[(source, target)]


    def add_edge(self, v, w, **kwargs):
        """
        Adds an edge to a network
        """

        # Add nodes if they don't exist
        self.add_node(v)
        self.add_node(w)

        e = (v, w)

        self.edges[e] = {**self.edges[e], **kwargs}

        if 'weight' not in self.edges[e]:
            self.edges[e]['weight'] = 1.0

        # update predecessor and successor lists
        self.successors[v].add(w)
        self.predecessors[w].add(v)
        if not self.directed:
            self.successors[w].add(v)
            self.predecessors[v].add(w)

        # update degrees and node weights
        if not self.directed:
            # update degree, in- and outweight
            self.nodes[v]['degree'] = len(self.successors[v])
            self.nodes[w]['degree'] = len(self.successors[w])

            S = [self.edges[(v,w)]['weight'] for w in self.successors[v]]
            if S:
                self.nodes[v]['outweight'] = sum(S)
                self.nodes[v]['inweight'] = self.nodes[v]['outweight']

            S = [self.edges[(v,w)]['weight'] for v in self.predecessors[w]]
            if S:
                self.nodes[w]['outweight'] = sum(S)
                self.nodes[w]['inweight'] = self.nodes[w]['outweight']
        else:
            self.nodes[v]['outdegree'] = len(self.successors[v])
            self.nodes[v]['indegree'] = len(self.predecessors[v])
            self.nodes[w]['outdegree'] = len(self.successors[w])
            self.nodes[w]['indegree'] = len(self.predecessors[w])

            # Note: Weights will be 0 for nodes with empty successors or predecessors. This is a
            # problem for higher-order networks, where the zero weight is assumed to be a vector
            # (0,0), Not updating weights in this case will ensure that we keep the initial value
            # of weights

            S = [self.edges[(v, x)]['weight'] for x in self.successors[v]]
            if S:
                self.nodes[v]['outweight'] = sum(S)
            S = [self.edges[(x, v)]['weight'] for x in self.predecessors[v]]
            if S:
                self.nodes[v]['inweight'] = sum(S)
            S = [self.edges[(w, x)]['weight'] for x in self.successors[w]]
            if S:
                self.nodes[w]['outweight'] = sum(S)
            S = [self.edges[(x, w)]['weight'] for x in self.predecessors[w]]
            if S:
                self.nodes[w]['inweight'] = sum(S)


    def find_nodes(self, select_node=lambda v: True):
        """
        Returns all nodes that satisfy a given condition. In the select_node 
        lambda function, node attributes can be accessed by calling v['attr']
        """
        return [n for n in self.nodes if select_node(self.nodes[n])]


    def find_edges(self, select_nodes=lambda v, w: True, select_edges=lambda e: True):
        """
        Returns all edges that satisfy a given condition. Edges can be selected based
        on attributes of the adjacent nodes as well as attributes of the edge. In the select_edges 
        lambda function, edge attributes can be accessed by calling e['attr']
        """
        return [e for e in self.edges if (select_nodes(self.nodes[e[0]], self.nodes[e[1]]) and select_edges(self.edges[e]))]


    def vcount(self):
        """ Returns the number of nodes """
        return len(self.nodes)


    def ecount(self):
        """ Returns the number of links """
        return len(self.edges)


    def total_edge_weight(self):
        """ Returns the sum of all edge weights """
        if self.edges:
            return _np.sum(e['weight'] for e in self.edges.values())
        return 0


    def node_to_name_map(self):
        """Returns a dictionary that can be used to map nodes to matrix/vector indices"""
        return {v: idx for idx, v in enumerate(self.nodes)}


    def adjacency_matrix(self, weighted=True, transposed=False):
        """Returns a sparse adjacency matrix of the higher-order network. Unless transposed
        is set to true, the entry corresponding to a directed link s->t is stored in row s and
        column t and can be accessed via A[s,t].

        Parameters
        ----------
        weighted: bool
            if set to False, the function returns a binary adjacency matrix.
            If set to True, adjacency matrix entries contain edge weights.
        transposed: bool
            whether to transpose the matrix or not.

        Returns
        -------
        numpy cooc matrix
        """
        row = []
        col = []
        data = []

        edgeC = self.ecount()
        if not self.directed:
            edgeC *= 2

        node_to_coord = self.node_to_name_map()
        
        for s, t in self.edges:
            row.append(node_to_coord[s])
            col.append(node_to_coord[t])
            if not self.directed:
                row.append(node_to_coord[t])
                col.append(node_to_coord[s])

        # create array with non-zero entries
        if not weighted:
            data = _np.ones(edgeC)
        else:
            data = _np.array([float(e['weight']) for e in self.edges.values()])
            if not self.directed:
                data = _np.column_stack((data, data)).flatten()

        shape = (self.vcount(), self.vcount())
        A = _sparse.coo_matrix((data, (row, col)), shape=shape).tocsr()

        if transposed:
            return A.transpose()
        return A


    def transition_matrix(self):
        """Returns a (transposed) transition matrix of a random walk process
        on the network

        Parameters
        ----------

        Returns
        -------

        """
        row = []
        col = []
        data = []

        # calculate weighted out-degrees of all nodes
        D = {n: self.nodes[n]['outweight'] for n in self.nodes}

        node_to_coord = self.node_to_name_map()

        for s, t in self.edges:
            # the following makes sure that we do not accidentally consider zero-weight
            # edges (automatically added by default_dic)
            weight = self.edges[(s, t)]['weight']
            if weight > 0:
                # add transition from s to t
                row.append(node_to_coord[t])
                col.append(node_to_coord[s])
                assert D[s] > 0, \
                    'Encountered zero out-weight or out-degree for node "{s}" ' \
                    'while weight of link ({s}, {t}) is non-zero.'.format(s=s, t=t)
                prob = weight / D[s]
                if prob < 0 or prob > 1:  # pragma: no cover
                    raise ValueError('Encountered transition probability {p} outside '
                                     '[0,1] range.'.format(p=prob))
                data.append(prob)

                # add transition from t to s for undirected network
                if not self.directed:
                    row.append(node_to_coord[s])
                    col.append(node_to_coord[t])
                    assert D[t] > 0, \
                    'Encountered zero out-degree for node "{t}" ' \
                    'while weight of link ({t}, {s}) is non-zero.'.format(s=s, t=t)
                    prob = weight / D[t]
                    if prob < 0 or prob > 1:  # pragma: no cover
                        raise ValueError('Encountered transition probability {p} outside '
                                        '[0,1] range.'.format(p=prob))
                    data.append(prob)

        data = _np.array(data)
        data = data.reshape(data.size, )

        shape = self.vcount(), self.vcount()
        return _sparse.coo_matrix((data, (row, col)), shape=shape).tocsr()


    def laplacian_matrix(self, weighted=False, transposed=False):
        """
        Returns the transposed normalized Laplacian matrix corresponding to the network.

        Parameters
        ----------

        Returns
        -------

        """
        if weighted:
            A = self.transition_matrix().transpose()
            D = _sparse.identity(self.vcount())
        else:
            A = self.adjacency_matrix(weighted=False)
            D = _sparse.diags(_np.array([float(self.nodes[v]['degree']) for v in self.nodes]))
        L = D - A
        if transposed:
            return L.transpose()
        return L


    @staticmethod
    def leading_eigenvector(A, normalized=True, lanczos_vecs=None, maxiter=None):
        """Compute normalized leading eigenvector of a given matrix A.

        Parameters
        ----------
        A:
            sparse matrix for which leading eigenvector will be computed
        normalized: bool
            whether or not to normalize, default is True
        lanczos_vecs: int
            number of Lanczos vectors to be used in the approximate
            calculation of eigenvectors and eigenvalues. This maps to the ncv parameter
            of scipy's underlying function eigs.
        maxiter: int
            scaling factor for the number of iterations to be used in the
            approximate calculation of eigenvectors and eigenvalues.

        Returns
        -------

        """
        if not _sparse.issparse(A):  # pragma: no cover
            raise TypeError("A must be a sparse matrix")

        # NOTE: ncv sets additional auxiliary eigenvectors that are computed
        # NOTE: in order to be more confident to find the one with the largest
        # NOTE: magnitude, see https://github.com/scipy/scipy/issues/4987
        if lanczos_vecs == None or maxiter == None:
            w, pi = _sla.eigs(A, k=1, which="LM")
        else:
            w, pi = _sla.eigs(A, k=1, which="LM", ncv=lanczos_vecs, maxiter=maxiter)
        pi = pi.reshape(pi.size, )
        if normalized:
            pi /= sum(pi)
        return pi


    def summary(self):
        """Returns a string containing basic summary statistics of this network instance
        """
        summary_fmt = (
            '{directed_str} network\n'
            '\n'
            'Nodes:\t\t\t\t{vcount}\n'
            'Links:\t\t\t\t{ecount}\n'
        )
        if self.directed:
            directed_str = 'Directed'
        else:
            directed_str = 'Undirected'
        summary = summary_fmt.format(directed_str=directed_str, vcount=self.vcount(), ecount=self.ecount())
        return summary

    def __str__(self):
        """Returns the default string representation of this graphical model instance"""
        return self.summary()

    def _to_html(self, width=800, height=800, clusters=None, sizes=None, template_file=None, **kwargs):
        import json
        import os
        from string import Template

        # prefix nodes starting with number
        def fix_node_name(v):
            if v[0].isdigit():
                return "n_" + v
            return v

        network_data = {
            'links': [
                {'source': fix_node_name(e[0]),
                 'target': fix_node_name(e[1]),
                 'value': 1} for e in self.edges.keys()
            ]
        }

        def get_cluster(v):
            if clusters == None or v not in clusters:
                return 'None'
            else:
                return clusters[v]

        def get_size(v):
            if sizes == None or v not in sizes:
                return 6
            else:
                return sizes[v]

        network_data['nodes'] = [{'id': fix_node_name(v), 'group': get_cluster(v), 'size': get_size(v)} for v in self.nodes]

        import string
        import random

        div_id = "".join(random.choice(string.ascii_letters) for x in range(8))

        if template_file is None:
            module_dir = os.path.dirname(os.path.realpath(__file__))
            html_dir = os.path.join(module_dir, os.path.pardir, 'html_templates')            
            template_file = os.path.join(html_dir, 'network.html') 

        with open(template_file) as f:
            html_str = f.read()

        if self.directed:
            directedness = 'true'
        else:
            directedness = 'false'

        default_args = {
            'network_data': json.dumps(network_data),
            'directed' : directedness,
            'width': width,
            'height': height,
            'div_id': div_id
        }

        # replace all placeholders in template
        html = Template(html_str).substitute({**default_args, **kwargs})

        return html


    def _repr_html_(self, clusters=None, sizes=None, template_file=None, **kwargs):
        """
        display an interactive D3 visualisation of the higher-order network in jupyter
        """
        from IPython.core.display import display, HTML
        display(HTML(self._to_html(clusters=clusters, sizes=sizes, template_file=template_file, **kwargs)))


    def write_html(self, filename, width=800, height=800, clusters=None, sizes=None, template_file=None, **kwargs):
        html = self._to_html(width=width, height=height, clusters=clusters, sizes=sizes, template_file=template_file, **kwargs)
        with open(filename, 'w+') as f:
            f.write(html)


def network_from_networkx(graph):
    """method to load a networkx graph into a pathpy.Network instance

    Parameters
    ----------
    garph

    Returns
    -------
    Network
    """
    try:
        import networkx as nx
    except ImportError:
        raise PathpyError("To load a network from networkx it must be installed")

    if isinstance(graph, nx.DiGraph):
        directed = True
    elif isinstance(graph, nx.Graph):
        directed = False
    else:
        raise PathpyNotImplemented("At the moment only DiGraph and Graph are supported.")

    net = Network(directed=directed)
    for node_id in graph.nodes:
        net.add_node(str(node_id), **graph.node[node_id])

    for edge in graph.edges:
        net.add_edge(str(edge[0]), str(edge[1]), **graph.edges[edge])

    return net


def network_to_networkx(network):
    """method to export a pathpy Network to a networkx compatible graph

    Parameters
    ----------
    network: Network

    Returns
    -------
    networkx Graph or DiGraph
    """
    # keys to exclude since they are handled differently in networkx
    excluded_node_props = {"degree", "inweight", "outweight", "indegree", "outdegree"}
    try:
        import networkx as nx
    except ImportError:
        raise PathpyError("To export a network to networkx it must be installed")

    directed = network.directed
    if directed:
        graph = nx.DiGraph()
    else:
        graph = nx.Graph()

    for node_id, node_props in network.nodes.items():
        valid_props = {k: v for k, v in node_props.items() if k not in excluded_node_props}
        graph.add_node(node_id, **valid_props)

    for edge, edge_props in network.edges.items():
        graph.add_edge(*edge, **edge_props)

    return graph


class UnorderedDict(dict):
    """A dictionary that applies an arbitrary key-altering
       function before accessing the keys

       Source: https://stackoverflow.com/questions/3387691/how-to-perfectly-override-a-dict
       """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.store = _co.defaultdict(dict)
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]

    def __contains__(self, item):
        return self.__keytransform__(item) in self.store

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __missing__(self, key):
        return {}

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def __keytransform__(self, key):
        return tuple(sorted(key))



from .dag import DAG
from .higher_order_network import HigherOrderNetwork
from .markov_sequence import MarkovSequence
from .multi_order_model import MultiOrderModel
from .paths import Paths
from .temporal_network import TemporalNetwork
from .network import Network
from .random_walk import random_walk




#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : tntp.py -- Convert tntp data to various other formats
# Author    : Juergen Hackl <hackl@ibi.baug.ethz.ch>
# Creation  : 2018-07-20
# Time-stamp: <Sam 2018-07-21 16:06 juergen>
#
# Copyright (c) 2018 Juergen Hackl <hackl@ibi.baug.ethz.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

from collections import defaultdict
import cnet as cn
from cnet import logger

log = logger(__name__)


class TNTPConverter(object):
    """A class to convert TNTP [1]_ data to various other formats.

    Parameters
    ----------
    filename : file or string, optional (default = None)
        File or filename to load. The file must be in a '.tntp' format.

    References
    ----------
    .. [1] https://github.com/bstabler/TransportationNetworks

    """

    def __init__(self, filename=None):
        self.filename = filename

    def trips(self, filename=None, prefix='', zfill=0):
        """Converts od trips to a dictionary format.

        Parameters
        ----------
        filename : file or string, optional (default = None)
            File or filename to load. The file must be in a '.tntp' format.

        prefix : string, optional (default = '')
            Adds a string value in front of the node id.

        zfill : integer, optional (default = 0)
            Fills the empty space in front of the node ids with zeros. The
            number indicate the digets filled up, e.g. `zfill = 2` transforms
           '1' to '01'.

        Returns
        -------
        trips : defaultdict
            Returns the od trips as a dictionary of dictionaries, where the
            first key defines the origin node of the flow, and the second key
            the destination node, the value is the traffic flow between origin
            and destination in vehicles per time unit.

        """
        # initialize variables
        trips = defaultdict(dict)
        zones = 0

        # load file content
        content = self._load_file(filename)

        # end function if no file was loaded
        if content is None:
            return None

        destinations = []
        # iterate through the content and get origin nodes
        for i, line in enumerate(content):

            # If text origin is observed or the end of the document reached,
            # do something.
            if 'Origin' in line or (i+1) == len(content):
                if destinations:
                    destinations = list(
                        filter(None, ''.join(destinations).split(';')))

                # get destination nodes and trips
                # in order to add them to the dict
                for d in destinations:
                    D = prefix+d.split(':')[0].strip(' \t\n\r').zfill(zfill)
                    V = float(d.split(':')[1].strip(' \t\n\r'))
                    trips[O][D] = V

                # check if the end of the file is reached,
                # otherwise go to next origin node
                if (i+1) != len(content):
                    # add new origin node
                    O = prefix+line.split('Origin',
                                          1)[1].split()[0].zfill(zfill)
                    # update temporal variables
                    zones += 1
                    destinations = []

            # add lines to temporal list
            if zones > 0 and 'Origin' not in line:
                destinations.append(line)

        return trips

    def network(self, filename=None, name='tntp', prefix='', zfill=0):
        """Generate a cnet network from tntp file.

        Parameters
        ----------
        filename : file or string, optional (default = None)
            File or filename to load. The file must be in a '.tntp' format.

        prefix : string or tuple of string, optional (default = '')
            Adds a string value in front of the node/edge id.

        zfill : integer, optional (default = 0)
            Fills the empty space in front of the node/edge ids with zeros. The
            number indicate the digets filled up, e.g. `zfill = 2` transforms
           '1' to '01'.

        Returns
        -------
        network : :py:class:`RoadNetwork`
            Returns a road network.

        """
        # initialize variables
        network = cn.RoadNetwork(name=name, directed=True)

        # load file content
        content = self._load_file(filename)

        # end function if no file was loaded
        if content is None:
            return None
        # check prefix
        if isinstance(prefix, tuple):
            prefix_n = prefix[0]
            prefix_e = prefix[1]
        else:
            prefix_n = prefix
            prefix_e = prefix

        edges = False

        for i, line in enumerate(content):
            if edges:
                id = prefix_e + str(i-j).zfill(zfill)
                e = line.split('\t')
                u = prefix_n + e[0].zfill(zfill)
                v = prefix_n + e[1].zfill(zfill)
                c = float(e[2])
                l = float(e[3])
                fft = float(e[4])
                ffs = l/fft
                a = float(e[5])
                b = float(e[6])
                sl = float(e[7])
                to = float(e[8])
                ty = float(e[9])

                edge = cn.RoadEdge(id, u, v, p1=(0, 0), p2=(0, 0), capacity=c,
                                   length=l, free_flow_speed=ffs, alpha=a,
                                   beta=b, speed_limit=sl, toll=to, type=ty)
                network.add_edge(edge)
            # set edges to True after reading the header.
            if '~' in line:
                j = i
                edges = True

        return network

    def _load_file(self, filename):
        """Function to check and lode the input file."""
        # check the input file name
        if self.filename is None and filename is None:
            log.warn('No file name was defined,'
                     'hence, no output was created!')
            return None
        elif filename is not None:
            self.filename = filename

        # load the file if the ending is correct
        if self.filename.endswith('.tntp'):
            # log.debug('Read tntp file.')
            with open(filename, 'r') as f:
                content = f.readlines()
                # remove ' \t\n\r' form the file
                content = [x.strip() for x in content]
                return content
        else:
            log.warn('The file must be in a ".tntp" format.'
                     'Please use correct file format. No output was created!')
            return None

        # =============================================================================
        # eof
        #
        # Local Variables:
        # mode: python
        # mode: linum
        # mode: auto-fill
        # fill-column: 80
        # End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : __init__.py
# Creation  : 21 May 2018
# Time-stamp: <Mon 2018-07-30 15:41 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : init file for the package
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

from .__about__ import(
    __title__,
    __version__,
    __author__,
    __email__,
    __copyright__,
    __license__,
    __maintainer__,
    __status__,
    __credits__
)

import logging


def logger(name, level='INFO'):
    """A function to generate logger for the modules."""
    # initialize new logger
    logger = logging.getLogger(name)
    # set logger level
    logger.setLevel(logging._nameToLevel[level])
    return logger


from .plot import Plot
from .layout import layout
plot = Plot()

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : layout.py -- Module to layout the network
# Author    : Juergen Hackl <hackl@ibi.baug.ethz.ch>
# Creation  : 2018-07-26
# Time-stamp: <Mon 2018-07-30 10:35 juergen>
#
# Copyright (c) 2018 Juergen Hackl <hackl@ibi.baug.ethz.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================
import numpy as np
from . import logger
from .exceptions import CnetError, CnetNotImplemented
log = logger(__name__)


def layout(network, **kwds):
    """Function to generate a layout for the network.

    This function genearates a layout configuration for the nodes in the
    network. Thereby, different layouts and options can be chosen. The layout
    function is directly included in the plot function or can be separately
    called.

    The layout function supports different network types and layout algorithm.
    Currently supported networks are:

    * 'cnet',
    * 'networkx',
    * 'igraph',
    * 'pathpy'
    * node/edge list

    Currently supported algorithms are:

    * Fruchterman-Reingold force-directed algorithm
    * Uniformly at random node positions

    The appearance of the layout can be modified by keyword arguments which will
    be explained in more detail below.

    Parameters
    ----------

    network : network object
        Network to be drawn. The network can be a 'cnet', 'networkx', 'igraph',
        'pathpy' object, or a tuple of a node list and edge list.

    kwds : keyword arguments, optional (default= no attributes)
        Attributes used to modify the appearance of the layout.
        For details see below.


    Keyword arguments used for the layout:

    **Layout:**

    NOTE: All layout arguments can be entered with or without 'layout_' at the
    beginning, e.g. 'layout_iterations' is equal to 'iterations'

    - ``layout`` : dict or string , optional (default = None)
      A dictionary with the node positions on a 2-dimensional plane. The
      key value of the dict represents the node id while the value
      represents a tuple of coordinates (e.g. n = (x,y)). The initial
      layout can be placed anywhere on the 2-dimensional plane.

      Instead of a dictionary, the algorithm used for the layout can be defined
      via a string value. Currently, supported are:

      * Random layout, where the nodes are uniformly at random placed in the
        unit square. This algorithm can be enabled with the keywords: 'Random',
        'random', 'rand', or None

      * Fruchterman-Reingold force-directed algorithm. In this algorithm, the
        nodes are represented by steel rings and the edges are springs between
        them. The attractive force is analogous to the spring force and the
        repulsive force is analogous to the electrical force. The basic idea is
        to minimize the energy of the system by moving the nodes and changing
        the forces between them. This algorithm can be enabled with the
        keywords: 'Fruchterman-Reingold', 'fruchterman_reingold', 'fr',
        'spring_layout', 'spring layout', 'FR'

        ==================== ==================================================
        Algorithms           Keywords
        ==================== ==================================================
        Random               Random, random, rand, None
        Fruchterman-Reingold Fruchterman-Reingold, fruchterman_reingold, fr
                             spring_layout, spring layout, FR
        ==================== ==================================================

    - ``force`` : float, optional (default = None)
      Optimal distance between nodes.  If None the distance is set to
      1/sqrt(n) where n is the number of nodes.  Increase this value to move
      nodes farther apart.

    - ``positions`` : dict or None  optional (default = None)
      Initial positions for nodes as a dictionary with node as keys and values
      as a coordinate list or tuple.  If None, then use random initial
      positions.

    - ``fixed`` : list or None, optional (default = None)
      Nodes to keep fixed at initial position.

    - ``iterations`` : int, optional (default = 50)
      Maximum number of iterations taken

    - ``threshold``: float, optional (default = 1e-4)
      Threshold for relative error in node position changes.  The iteration
      stops if the error is below this threshold.

    - ``weight`` : string or None, optional (default = None)
      The edge attribute that holds the numerical value used for the edge
      weight.  If None, then all edge weights are 1.

    - ``dimension`` : int, optional (default = 2)
      Dimension of layout. Currently, only plots in 2 dimension are supported.

    - ``seed`` : int or None, optional (default = None)
      Set the random state for deterministic node layouts. If int, `seed` is
      the seed used by the random number generator, if None, the a random seed
      by created by the numpy random number generator is used.

    In the layout style dictionary multiple keywords can be used to address
    attributes. These keywords will be converted to an unique key word,
    used in the remaining code.

    ========= =================================
    keys      other valid keys
    ========= =================================
    fixed     fixed_nodes, fixed_vertices,
              fixed_n, fixed_v
    positions initial_positions, node_positions
              vertex_positions, n_positions,
              v_positions
    ========= =================================

    Examples
    --------

    For illustration purpose a similar network as in the python-igrap tutorial
    is used. Instead of igraph, the cnet module is used for creating the
    network.

    Create an empty network object, and add some edges.

    >>> net = Network(name = 'my tikz test network',directed=True)
    >>> net.add_edges_from([('ab','a','b'), ('ac','a','c'), ('cd','c','d'),
    >>>                     ('de','d','e'), ('ec','e','c'), ('cf','c','f'),
    >>>                     ('fa','f','a'), ('fg','f','g'),('gg','g','g'),
    >>>                     ('gd','g','d')])

    Now a layout can be generated:

    >>> layout(net)
    {'b': array([0.88878309, 0.15685131]), 'd': array([0.4659341 , 0.79839535]),
    'c': array([0.60386662, 0.40727962]), 'e': array([0.71073353, 0.65608203]),
    'g': array([0.42663927, 0.47412449]), 'f': array([0.48759769, 0.86787594]),
    'a': array([0.84154488, 0.1633732 ])}

    Per default, the node positions are assigned uniform random. In order to
    create a layout, the layout methods of the packages can be used, or the
    position of the nodes can be directly assigned, in form of a dictionary,
    where the key is the node id and the value is a tuple of the node position
    in x and y.

    Let us generate a force directed layout (e.g. Fruchterman-Reingold):

    >>> layout(net, layout='fr')
    {'g': array([-0.77646408,  1.71291126]), 'c': array([-0.18639655,0.96232326]),
    'f': array([0.33394308, 0.93778681]), 'e': array([0.09740098, 1.28511973]),
    'a': array([1.37933158, 0.23171857]), 'b': array([ 2.93561876,-0.46183461]),
    'd': array([-0.29329793,  1.48971303])}

    Note, instead of the command ``fr`` also the command
    ``Fruchterman-Reingold`` or any other command mentioned above can be
    used. For more information see table above.

    In order to keep the properties of the layout for your network separate from
    the network itself, you can simply set up a Python dictionary containing the
    keyword arguments you would pass to :py:meth:`layout` and then use the
    double asterisk (**) operator to pass your specific layout attributes to
    :py:meth:`layout`:

    >>> layout_style = {}
    >>> layout_style['layout'] = 'Fruchterman-Reingold'
    >>> layout_style['seed'] = 1
    >>> layout_style['iterations'] = 100
    >>> layout(net,**layout_style)
    {'d': array([-0.31778276, 1.78246882]), 'f': array([-0.8603259, 0.82328291]),
    'c': array([-0.4423771 , 1.21203895]), 'e': array([-0.79934355, 1.49000119]),
    'g': array([0.43694799, 1.51428788]), 'a': array([-2.15517293, 0.23948823]),
    'b': array([-3.84803812, -0.71628417])}

    """
    # initialize variables
    _weight = kwds.get('weight', None)
    if _weight is None:
        _weight = kwds.get('layout_weight', None)

    # check type of network
    if 'cnet' in str(type(network)):
        # log.debug('The network is of type "cnet".')
        nodes = list(network.nodes)
        adjacency_matrix = network.adjacency_matrix(weight=_weight)

    elif 'networkx' in str(type(network)):
        # log.debug('The network is of type "networkx".')
        nodes = list(network.nodes())
        import networkx as nx
        adjacency_matrix = nx.adjacency_matrix(network, weight=_weight)
    elif 'igraph' in str(type(network)):
        # log.debug('The network is of type "igraph".')
        nodes = list(range(len(network.vs)))
        from scipy.sparse import coo_matrix
        A = np.array(network.get_adjacency(attribute=_weight).data)
        adjacency_matrix = coo_matrix(A)
    elif 'pathpy' in str(type(network)):
        # log.debug('The network is of type "pathpy".')
        nodes = list(network.nodes)
        if _weight is not None:
            _w = True
        else:
            _w = False
        adjacency_matrix = network.adjacency_matrix(weighted=_w)
    # elif isinstance(network, tuple):
    #     # log.debug('The network is of type "list".')
    #     nodes = network[0]
    #     from collections import OrderedDict
    #     edges = OrderedDict()
    #     for e in network[1]:
    #         edges[e] = e

    else:
        log.error('Type of the network could not be determined.'
                  ' Currently only "cnet", "networkx","igraph", "pathpy"'
                  ' and "node/edge list" is supported!')
        raise CnetNotImplemented

    # create layout class
    layout = Layout(nodes, adjacency_matrix, **kwds)
    # return the layout
    return layout.generate_layout()


class Layout(object):
    """Default class to create layouts

    The :py:class:`Layout` class is used to generate node a layout drawer and
    return the calculated node positions as a dictionary, where the keywords
    represents the node ids and the values represents a two dimensional tuple
    with the x and y coordinates for the associated nodes.

    Parameters
    ----------
    nodes : list with node ids
        The list contain a list of unique node ids.

    attr : keyword arguments, optional (default = no attributes)
        Attributes to add to node as key=value pairs.
        See also :py:meth:`layout`

    See Also
    --------
    layout

    """

    def __init__(self, nodes, adjacency_matrix, **attr):
        """Initialize the Layout class

        The :py:class:`Layout` class is used to generate node a layout drawer
        and return the calculated node positions as a dictionary, where the
        keywords represents the node ids and the values represents a two
        dimensional tuple with the x and y coordinates for the associated nodes.

        Parameters
        ----------
        nodes : list with node ids
            The list contain a list of unique node ids.

        attr : keyword arguments, optional (default = no attributes)
            Attributes to add to node as key=value pairs.
            See also :py:meth:`layout`

        """

        # initialize variables
        self.nodes = nodes
        self.adjacency_matrix = adjacency_matrix

        # rename the attributes
        attr = self.rename_attributes(**attr)

        # options for the layouts
        self.layout_type = attr.get('layout', None)
        self.k = attr.get('force', None,)
        self.fixed = attr.get('fixed', None)
        self.iterations = attr.get('iterations', 50)
        self.threshold = attr.get('threshold', 1e-4)
        self.weight = attr.get('weight', None)
        self.dimension = attr.get('dimension', 2)
        self.seed = attr.get('seed', None)
        self.positions = attr.get('positions', None)

        # TODO: allow also higher dimensional layouts
        if self.dimension != 2:
            log.warning('Currently only plots with dimension 2 are supported!')
            self.dimension = 2

    @staticmethod
    def rename_attributes(**kwds):
        """Rename layout attributes.

        In the style dictionary multiple keywords can be used to address
        attributes. These keywords will be converted to an unique key word,
        used in the remaining code.

        ========= =================================
        keys      other valid keys
        ========= =================================
        fixed     fixed_nodes, fixed_vertices,
                  fixed_n, fixed_v
        positions initial_positions, node_positions
                  vertex_positions, n_positions,
                  v_positions
        ========= =================================

        """
        names = {'fixed': ['fixed_nodes', 'fixed_vertices',
                           'fixed_v', 'fixed_n'],
                 'positions': ['initial_positions', 'node_positions',
                               'vertex_positions', 'n_positions',
                               'v_positions'],
                 'layout_': ['layout_'],
                 }

        _kwds = {}
        del_keys = []
        for key, value in kwds.items():
            for attr, name_list in names.items():
                for name in name_list:
                    if name in key and name[0] == key[0]:
                        _kwds[key.replace(name, attr).replace(
                            'layout_', '')] = value
                        del_keys.append(key)
                        break
        # remove the replaced keys from the dict
        for key in del_keys:
            del kwds[key]

        return {**_kwds, **kwds}

    def generate_layout(self):
        """Function to pick and generate the right layout."""
        # method names
        names_rand = ['Random', 'random', 'rand', None]
        names_fr = ['Fruchterman-Reingold', 'fruchterman_reingold', 'fr',
                    'spring_layout', 'spring layout', 'FR']
        # check which layout should be plotted
        if self.layout_type in names_rand:
            self.layout = self.random()
        elif self.layout_type in names_fr:
            self.layout = self.fruchterman_reingold()

        # print(self.layout)
        return self.layout

    def random(self):
        """Position nodes uniformly at random in the unit square.

        For every node, a position is generated by choosing each of dimension
        coordinates uniformly at random on the interval [0.0, 1.0).

        This algorithm can be enabled with the keywords: 'Random',
        'random', 'rand', or None

        NumPy (http://scipy.org) is required for this function.

        **Keyword arguments used for the layout:**

        - ``dimension`` : int, optional (default = 2)
          Dimension of layout. Currently, only plots in 2 dimension are supported.

        - ``seed`` : int or None, optional (default = None)
          Set the random state for deterministic node layouts. If int, `seed` is
          the seed used by the random number generator, if None, the a random
          seed by created by the numpy random number generator is used.

        Returns
        -------
        layout : dict
            A dictionary of positions keyed by node

        """
        np.random.seed(self.seed)
        layout = np.random.rand(len(self.nodes), self.dimension)
        return dict(zip(self.nodes, layout))

    def fruchterman_reingold(self):
        """Position nodes using Fruchterman-Reingold force-directed algorithm.

        In this algorithm, the nodes are represented by steel rings and the
        edges are springs between them. The attractive force is analogous to the
        spring force and the repulsive force is analogous to the electrical
        force. The basic idea is to minimize the energy of the system by moving
        the nodes and changing the forces between them.

        This algorithm can be enabled with the keywords: 'Fruchterman-Reingold',
        'fruchterman_reingold', 'fr', 'spring_layout', 'spring layout', 'FR'

        **Keyword arguments used for the layout:**

        - ``force`` : float, optional (default = None)
          Optimal distance between nodes.  If None the distance is set to
          1/sqrt(n) where n is the number of nodes.  Increase this value to move
          nodes farther apart.

        - ``positions`` : dict or None  optional (default = None)
          Initial positions for nodes as a dictionary with node as keys and values
          as a coordinate list or tuple.  If None, then use random initial
          positions.

        - ``fixed`` : list or None, optional (default = None)
          Nodes to keep fixed at initial position.

        - ``iterations`` : int, optional (default = 50)
          Maximum number of iterations taken

        - ``threshold``: float, optional (default = 1e-4)
          Threshold for relative error in node position changes.  The iteration
          stops if the error is below this threshold.

        - ``weight`` : string or None, optional (default = None)
          The edge attribute that holds the numerical value used for the edge
          weight.  If None, then all edge weights are 1.

        - ``dimension`` : int, optional (default = 2)
          Dimension of layout. Currently, only plots in 2 dimension are supported.

        - ``seed`` : int or None, optional (default = None)
          Set the random state for deterministic node layouts. If int, `seed` is
          the seed used by the random number generator, if None, the a random seed
          by created by the numpy random number generator is used.

        Returns
        -------
        layout : dict
            A dictionary of positions keyed by node

        """

        # convert adjacency matrix
        self.adjacency_matrix = self.adjacency_matrix.astype(float)

        if self.fixed is not None:
            self.fixed = np.asarray([self.nodes.index(v) for v in self.fixed])

        if self.positions is not None:
            # Determine size of existing domain to adjust initial positions
            _size = max(coord for t in layout.values() for coord in t)
            if _size == 0:
                _size = 1
            np.random.seed(self.seed)
            self.layout = np.random.rand(
                len(self.nodes), self.dimension) * _size

            for i, n in enumerate(self.nodes):
                if n in self.positions:
                    self.layout[i] = np.asarray(self.positions[n])
        else:
            self.layout = None

        if self.k is None and self.fixed is not None:
            # We must adjust k by domain size for layouts not near 1x1
            self.k = _size / np.sqrt(len(self.nodes))

        try:
            # Sparse matrix
            if len(self.nodes) < 500:  # sparse solver for large graphs
                raise ValueError
            layout = self._sparse_fruchterman_reingold()
        except:
            layout = self._fruchterman_reingold()

        layout = dict(zip(self.nodes, layout))

        return layout

    def _fruchterman_reingold(self):
        """Fruchterman-Reingold algorithm for dense matrices.

        This algorithm is based on the Fruchterman-Reingold algorithm provided
        by networkx. (Copyright (C) 2004-2018 by Aric Hagberg <hagberg@lanl.gov>
        Dan Schult <dschult@colgate.edu> Pieter Swart <swart@lanl.gov> Richard
        Penney <rwpenney@users.sourceforge.net> All rights reserved. BSD
        license.)

        """
        A = self.adjacency_matrix.todense()
        k = self.k
        try:
            _n, _ = A.shape
        except AttributeError:
            log.error('Fruchterman-Reingold algorithm needs an adjacency '
                      'matrix as input')
            raise CnetError

        # make sure we have an array instead of a matrix
        A = np.asarray(A)

        if self.layout is None:
            # random initial positions
            np.random.seed(self.seed)
            layout = np.asarray(np.random.rand(
                _n, self.dimension), dtype=A.dtype)
        else:
            # make sure positions are of same type as matrix
            layout = self.layout.astype(A.dtype)

        # optimal distance between nodes
        if k is None:
            k = np.sqrt(1.0 / _n)
        # the initial "temperature"  is about .1 of domain area (=1x1)
        # this is the largest step allowed in the dynamics.
        # We need to calculate this in case our fixed positions force our domain
        # to be much bigger than 1x1
        t = max(max(layout.T[0]) - min(layout.T[0]),
                max(layout.T[1]) - min(layout.T[1])) * 0.1
        # simple cooling scheme.
        # linearly step down by dt on each iteration so last iteration is size dt.
        dt = t / float(self.iterations + 1)
        delta = np.zeros(
            (layout.shape[0], layout.shape[0], layout.shape[1]), dtype=A.dtype)
        # the inscrutable (but fast) version
        # this is still O(V^2)
        # could use multilevel methods to speed this up significantly
        for iteration in range(self.iterations):
            # matrix of difference between points
            delta = layout[:, np.newaxis, :] - layout[np.newaxis, :, :]
            # distance between points
            distance = np.linalg.norm(delta, axis=-1)
            # enforce minimum distance of 0.01
            np.clip(distance, 0.01, None, out=distance)
            # displacement "force"
            displacement = np.einsum('ijk,ij->ik',
                                     delta,
                                     (k * k / distance**2 - A * distance / k))
            # update layoutitions
            length = np.linalg.norm(displacement, axis=-1)
            length = np.where(length < 0.01, 0.1, length)
            delta_layout = np.einsum('ij,i->ij', displacement, t / length)
            if self.fixed is not None:
                # don't change positions of fixed nodes
                delta_layout[self.fixed] = 0.0
            layout += delta_layout
            # cool temperature
            t -= dt
            error = np.linalg.norm(delta_layout) / _n
            if error < self.threshold:
                break
        return layout

    def _sparse_fruchterman_reingold(self):
        """Fruchterman-Reingold algorithm for sparse matrices.

        This algorithm is based on the Fruchterman-Reingold algorithm provided
        by networkx. (Copyright (C) 2004-2018 by Aric Hagberg <hagberg@lanl.gov>
        Dan Schult <dschult@colgate.edu> Pieter Swart <swart@lanl.gov> Richard
        Penney <rwpenney@users.sourceforge.net> All rights reserved. BSD
        license.)

        """
        A = self.adjacency_matrix
        k = self.k
        try:
            _n, _ = A.shape
        except AttributeError:
            log.error('Fruchterman-Reingold algorithm needs an adjacency '
                      'matrix as input')
            raise CnetError
        try:
            from scipy.sparse import spdiags, coo_matrix
        except ImportError:
            log.error('The sparse Fruchterman-Reingold algorithm needs the '
                      'scipy package: http://scipy.org/')
            raise ImportError
        # make sure we have a LIst of Lists representation
        try:
            A = A.tolil()
        except:
            A = (coo_matrix(A)).tolil()

        if self.layout is None:
            # random initial positions
            np.random.seed(self.seed)
            layout = np.asarray(np.random.rand(
                _n, self.dimension), dtype=A.dtype)
        else:
            # make sure positions are of same type as matrix
            layout = layout.astype(A.dtype)

        # no fixed nodes
        if self.fixed is None:
            self.fixed = []

        # optimal distance between nodes
        if k is None:
            k = np.sqrt(1.0 / _n)
        # the initial "temperature"  is about .1 of domain area (=1x1)
        # this is the largest step allowed in the dynamics.
        t = max(max(layout.T[0]) - min(layout.T[0]),
                max(layout.T[1]) - min(layout.T[1])) * 0.1
        # simple cooling scheme.
        # linearly step down by dt on each iteration so last iteration is size dt.
        dt = t / float(self.iterations + 1)

        displacement = np.zeros((self.dimension, _n))
        for iteration in range(self.iterations):
            displacement *= 0
            # loop over rows
            for i in range(A.shape[0]):
                if i in self.fixed:
                    continue
                # difference between this row's node position and all others
                delta = (layout[i] - layout).T
                # distance between points
                distance = np.sqrt((delta**2).sum(axis=0))
                # enforce minimum distance of 0.01
                distance = np.where(distance < 0.01, 0.01, distance)
                # the adjacency matrix row
                Ai = np.asarray(A.getrowview(i).toarray())
                # displacement "force"
                displacement[:, i] +=\
                    (delta * (k * k / distance**2 - Ai * distance / k)).sum(axis=1)
            # update positions
            length = np.sqrt((displacement**2).sum(axis=0))
            length = np.where(length < 0.01, 0.1, length)
            delta_layout = (displacement * t / length).T
            layout += delta_layout
            # cool temperature
            t -= dt
            err = np.linalg.norm(delta_layout) / _n
            if err < self.threshold:
                break
        return layout


# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : higher_order_network.py • cnet -- Basic classes for path networks
# Author    : Juergen Hackl <hackl@ibi.baug.ethz.ch>
# Creation  : 2018-07-25
# Time-stamp: <Mit 2018-07-25 16:00 juergen>
#
# Copyright (c) 2018 Juergen Hackl <hackl@ibi.baug.ethz.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

from cnet.classes.network import Node, Edge, Network
from cnet.classes.paths import Path, Paths
from cnet.utils.exceptions import CnetError, CnetNotImplemented
from cnet import config, logger
log = logger(__name__)


class PathNetwork(Network):
    """Base class for path networks.

    """

    def __init__(self, directed=True, **attr):
        """Initialize a network with direction, name and attributes.

        Parameters
        ----------
        directed : Boole, optional (default = True)
            Specifies if a network contains directed edges, i.e u->v or
            undirected edges i.d. u<->v. Per default the network is assumed
            to be directed.

        name : string, optional (default='')
            An optional name for the graph.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to graph as key=value pairs.

        """

        # initialize parent class of the network
        Network.__init__(self, directed=directed, **attr)

        # Classes of the Node and Edge objects
        # TODO: Probably there is a better solution to have different Node and
        # Edge classes for different Network sub classes
        self._node_class()
        self._edge_class()

    def _node_class(self):
        """Internal function to assign different Node classes."""
        self.NodeClass = PathNode

    def _edge_class(self):
        """Internal function to assign different Edge classes."""
        self.EdgeClass = PathEdge


class PathEdge(Edge):
    """Base class for an edge.

    Parameters
    ----------
    e : edge id
        The parameter e is the identifier (id) for the edge. Every edge should
        have a unique id. The id is converted to a string value and is used as a
        key value for all dict which saving edge objects.

    u : node id or py:class:`PathNode`
        This parameter defines the origin of the edge (if directed), i.e. u->v.
        A node id can be entered, in this case a new Node will be created
        (see py:class:`PathNode`), also an existing node can be used, here the
        attributes and properties of the node objects are used.

    v : node id or py:class:`PathNode`
        This parameter defines the destination of the edge (if directed)
        i.e. u->v. A node id can be entered, in this case a new Node will be
        created (see py:class:`PathNode`), also an existing node can be used,
        here the attributes and properties of the node objects are used.

    p1 : :py:class:`Path` object or :py:class:`Paths` object
        The path(s) associated with the node u. If no path is defined an empty
        path will be assigned to the node.

    p2 : :py:class:`Path` object or :py:class:`Paths` object
        The path(s) associated with the node v. If no path is defined an empty
        path will be assigned to the node.

    separator : string, optional (default = '|')
        The separator used to separate the nodes ids if the edge id is
        generated automatically (see edge id above).

    Examples
    --------

    """

    def __init__(self, id, u, v, **attr):
        """Initialize the edge object.

        Parameters
        ----------
        e : edge id
            The parameter e is the identifier (id) for the edge. Every edge
            should have a unique id. The id is converted to a string value and
            is used as a key value for all dict which saving edge objects.

        u : node id or py:class:`PathNode`
            This parameter defines the origin of the edge (if directed),
            i.e. u->v.  A node id can be entered, in this case a new Node will
            be created (see py:class:`PathNode`), also an existing node can be
            used, here the attributes and properties of the node objects are
            used.

        v : node id or py:class:`PathNode`
            This parameter defines the destination of the edge (if directed)
            i.e. u->v. A node id can be entered, in this case a new Node will be
            created (see py:class:`PathNode`), also an existing node can be
            used, here the attributes and properties of the node objects are
            used.

        p1 : :py:class:`Path` object or :py:class:`Paths` object
            The path(s) associated with the node u. If no path is defined an
            empty path will be assigned to the node.

        p2 : :py:class:`Path` object or :py:class:`Paths` object
            The path(s) associated with the node v. If no path is defined an
            empty path will be assigned to the node.

        separator : string, optional (default = '|')
            The separator used to separate the nodes ids if the edge id is
            generated automatically (see edge id above).

        """
        # Class of the Node object
        # TODO: Probably there is a better solution to have different Node
        # classes for different Edges sub classes
        self._node_class()

        # check the type of node object and if additional paths are defined
        if not isinstance(u, self.NodeClass) and 'p1' in attr:
            u = self.NodeClass(u, path=attr['p1'])
        if not isinstance(v, self.NodeClass) and 'p2' in attr:
            v = self.NodeClass(v, path=attr['p2'])

        # Create edge id based on the node ids
        if id is None:
            separator = attr.get('separator', '|')
            if isinstance(u, str):
                _u = u
            elif isinstance(u, self.NodeClass) or isinstance(u, Path):
                _u = u.id
            elif isinstance(u, Paths):
                if u.name == '' or u.name is None:
                    log.error('A unique node id has to be defined for node '
                              '{}!'.format(u))
                    raise CnetError
                else:
                    _u = u.name
            else:
                log.error('No valid node id for node {} was found'.format(u))
                raise CnetError

            if isinstance(v, str):
                _v = v
            elif isinstance(u, self.NodeClass) or isinstance(v, Path):
                _v = v.id
            elif isinstance(v, Paths):
                if v.name == '' or v.name is None:
                    log.error('A unique node id has to be defined for node '
                              '{}!'.format(v))
                    raise CnetError
                else:
                    _v = v.name
            else:
                log.error('No valid node id for node {} was found'.format(v))
                raise CnetError

            id = _u + separator + _v

        # initialize the parent edge class
        Edge.__init__(self, id, u, v, **attr)

    def _node_class(self):
        """Internal function to assign different Node classes to the edge"""
        self.NodeClass = PathNode

    def common_paths(self):
        """Returns the common paths between u and v."""
        return self.u.paths.intersection(self.v.paths)

    def common_subpaths(self, min_length=None, max_length=None, include_path=False):
        """Returns a paths object with a set of sub paths of u and v.

        Parameters
        ----------

        min_length : int, optional (default = None)
            Parameter which defines the minimum length of the sub paths. This
            parameter has to be smaller then the maximum length parameter.

        max_length : int, optional (default = None)
            Parameter which defines the maximum  length of the sub paths. This
            parameter has to be greater then the minimum length parameter. If
            the parameter is also greater then the maximum length of the path,
            the maximum path length is used instead.

        include_path : Boole, optional (default = Flase)
            If this option is enabled also the current path is added as a
            sub path of it self.

        Returns
        -------
        subpaths : Paths
            Returns a paths object containing a set of the sub paths fulfilling
            the length criteria.

        See Also
        --------
        :py:method:`Path.subpaths`
        """
        _subpaths_u = self.u.subpaths(min_length=min_length,
                                      max_length=max_length,
                                      include_path=include_path)
        _subpaths_v = self.v.subpaths(min_length=min_length,
                                      max_length=max_length,
                                      include_path=include_path)
        return _subpaths_u.intersection(_subpaths_v)


class PathNode(Node):
    """Base class for a path node.

    Thereby a path node represents one or multiple paths.

    Parameters
    ----------
    u : node id, :py:class:`Path` object or :py:class:`Paths` object
        The parameter u is the identifier (id) for the node. Every node should
        have a unique id. The id is converted to a string value and is used as a
        key value for all dict which saving node objects.  The path name is used
        as node id, if only a :py:class:`Path` object is given.  If only a
        :py:class:`Paths` object is given a valid name has to be defined.
        ATTENTION: IT IS RECOMMENDED TO USE A UNIQUE NODE ID INSTEAD OF PATHS!

    path : :py:class:`Path` object or :py:class:`Paths` object
        The path(s) associated with the node.

    Examples
    --------


    """

    def __init__(self, u=None, path=None, **attr):
        """Initialize the node object.

        Parameters
        ----------
        u : node id, :py:class:`Path` object or :py:class:`Paths` object
            The parameter u is the identifier (id) for the node. Every node
            should have a unique id. The id is converted to a string value and
            is used as a key value for all dict which saving node objects.  The
            path name is used as node id, if only a :py:class:`Path` object is
            given.  If only a :py:class:`Paths` object is given a valid name has
            to be defined.  ATTENTION: IT IS RECOMMENDED TO USE A UNIQUE NODE ID
            INSTEAD OF PATHS!

        path : :py:class:`Path` object or :py:class:`Paths` object
            The path(s) associated with the node.

        """
        # initialize variables
        _path = None
        _paths = None
        # check if path or paths are defined
        # if no node id or path is defined return an error message
        if u is None and path is None:
            log.error('Either the node id a single "path" or a list of "paths"'
                      ' must be defined!')
            raise CnetError
        # if no path is defined print a warning and add an empty path object
        elif isinstance(u, str) and path is None:
            log.warn('Either a single "path" or a list of "paths" must be '
                     'defined. Neither was found. Therefore an empty path was'
                     ' created for node {}.'.format(u))
        # if u is a path object, the node id will be the path id
        elif isinstance(u, Path):
            _path = u
            u = u.id
        # if u is a Paths object with valid name,
        # the node id will be the name of the Paths list
        elif isinstance(u, Paths):
            if u.name == '' or u.name is None:
                log.error('A unique path node id has to be defined!')
                raise CnetError
            else:
                _paths = u
                u = u.name
        # if node id and path is defined in the right format,
        # add them tho the node
        elif isinstance(u, str) and isinstance(path, Path):
            _path = path
        # if node id and paths are defined in the right format,
        # add them to the node
        elif isinstance(u, str) and \
                (isinstance(path, Paths) or isinstance(path, list)):
            if isinstance(path, Paths):
                _paths = path
            elif isinstance(path, list):
                _paths = Paths(path)
        # if non of the cases above a valid return an error message
        else:
            log.error('A PathNode must be defined with an id and a Path or '
                      ' Paths object!')
            raise CnetError

        # assign path and paths to the node object
        if _paths is None:
            self.paths = Paths([_path], name=u)
        else:
            self.paths = _paths
        self.path = _path

        # initialize parent class
        Node.__init__(self, u, **attr)

    def subpaths(self, min_length=None, max_length=None, include_path=False):
        """Returns a paths object with a set of sub paths.

        Parameters
        ----------

        min_length : int, optional (default = None)
            Parameter which defines the minimum length of the sub paths. This
            parameter has to be smaller then the maximum length parameter.

        max_length : int, optional (default = None)
            Parameter which defines the maximum  length of the sub paths. This
            parameter has to be greater then the minimum length parameter. If
            the parameter is also greater then the maximum length of the path,
            the maximum path length is used instead.

        include_path : Boole, optional (default = Flase)
            If this option is enabled also the current path is added as a
            sub path of it self.

        Returns
        -------
        subpaths : Paths
            Returns a paths object containing a set of the sub paths fulfilling
            the length criteria.

        See Also
        --------
        :py:method:`Path.subpaths`
        """
        _subpaths = set()
        for path in self.paths:
            for subpath in path.subpaths(min_length=min_length,
                                         max_length=max_length,
                                         include_path=include_path):
                _subpaths.add(subpath)
        subpaths = Paths(name='sub paths of node '+self.id)
        for path in _subpaths:
            subpaths.add_path(path)

        return subpaths
# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : __init__.py
# Creation  : 25 May 2018
# Time-stamp: <Son 2018-07-22 12:52 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : init file
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

#from .shortest_path import shortest_path, k_shortest_paths, ksp
#from .traffic_assignment import msa, msa_fast
#from .markov import estimate_transition_matrix, cost_matrix


# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

import collections as _co
import random

import numpy as _np

from pathpy.classes.network import Network
from pathpy.classes.paths import Paths


def random_walk(network, l, n=1):
    """
    Generates n paths of a random walker in the given network
    and returns them as a paths object. 
    Each path has a length of l steps   
    Parameters
    ----------
    network: pathpy.Network
    int: l
    int: n
    """

    p = Paths()
    T = network.transition_matrix().todense().transpose()
    idx_map = network.node_to_name_map()
    nodes = _np.array([v for v in network.nodes])

    for i in range(n):
        # choose random start node
        path = (random.choice(nodes),)
        for j in range(l):
            # get transition probability vector T[idx ->  . ]
            prob = _np.array(T[idx_map[path[-1]],:])[0,:]
            nz = prob.nonzero()[0]
            # make one random transition
            if nz.shape[0]>0:
                next_node = _np.random.choice(a=nodes[nz], p=prob[nz])
                # add node to path
                path = path + (next_node,)
            else:
                print('no neighbor')
                break
        p.add_path_tuple(path)
    return p




# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

from collections import defaultdict
import sys

from pathpy import Paths
from pathpy import DAG
from pathpy.path_extraction import paths_from_dag
from pathpy.utils import Log


def paths_from_temporal_network(tempnet, delta=1, max_length=sys.maxsize,
                                max_subpath_length=sys.maxsize):
    """
    Calculates the frequency of all time-respecting paths up to maximum length
    of maxLength, assuming a maximum temporal distance of delta between consecutive
    time-stamped links on a path. This (static) method returns an instance of the
    class Paths, which can subsequently be used to generate higher-order network
    representations based on the path statistics.

    Parameters
    ----------
    tempnet : pathpy.TemporalNetwork
        TemporalNetwork to extract the path from
    delta : int
        Indicates the maximum temporal distance up to which time-stamped
        links will be considered to contribute to time-respecting paths.
        For (u, v;3) and (v,w;7) a time-respecting path (u,v)->(v,w) will be inferred
        for all 0 < delta <= 4, while no time-respecting path will be inferred for all
        delta > 4. If the max time diff is not set specifically, the default value of
        delta=1 will be used, meaning that a time-respecting path u -> v -> w will
        only be inferred if there are *directly consecutive* time-stamped links
        (u,v;t) (v,w;t+1). Every time-stamped edge is further considered a path of
        length one, i.e. for maxLength=1 this function will simply return the
        statistics of time-stamped edges.
    max_length : int
        Indicates the maximum length up to which time-respecting paths should be
        calculated, which can be limited due to computational efficiency.
        A value of k will generate all time-respecting paths consisting of up to k
        time-stamped links. Note that generating a multi-order model with a maximum
        order of k requires to extract time-respecting paths with *at least* length k.
        If a limitation of the maxLength is not required for computational reasons,
        this parameter should not be set (as it will change the statistics of paths)
    max_subpath_length : int
        This can be used to limit the calculation of sub path statistics to a given
        maximum length. This is useful, as the statistics of sub paths of length k
        are only needed to fit a higher-order model with order k. Hence, if we know
        that the model selection is limited to a given maximum order K, we can safely
        set the maximum sub path length to K. By default, sub paths of any length
        will be calculated. Note that, independent of the sub path calculation
        longest path of any length will be considered in the likelihood calculation!

    Returns
    -------
    Paths

    """

    if max_length == sys.maxsize:  # pragma: no cover
        Log.add('Extracting time-respecting paths for delta = ' + str(delta) + ' ...')
    else:  # pragma: no cover
        Log.add('Extracting time-respecting paths up to length ' + str(max_length) +
                 ' for delta = ' + str(delta) + ' ...')

    # for dictionary p.paths paths[k] contains a list of all
    # time-respecting paths p with length k and paths[k][p] contains
    # a two-dimensional counter whose first component counts the number of
    # occurrences of p as subpath of a longer path and whose second component counts
    # the number of occurrences of p as "real" path
    p = Paths()

    p.max_subpath_length = max_subpath_length
    # a dictionary containing paths that can still be extended
    # by future time-stamped links
    # candidates[t][v] is a set of paths which end at time t in node v
    candidates = defaultdict(lambda: defaultdict(lambda: set()))

    # Note that here we only extract **longest** time-respecting
    # paths, since we will use later the expandSubpaths function to calculate statistics
    # of shorter paths

    # set of longest time-respecting paths (i.e. those paths which are
    # NOT sub path of a longer time-respecting path)
    longest_paths = set()

    # loop over all time stamps t of edges
    for t in tempnet.ordered_times:

        for e in tempnet.time[t]:
            # assume that this edge is the root of a longest time-respecting path
            root = True

            # check whether this edge extends existing candidates
            for t_prev in list(candidates):
                # time stamp of candidate has to be in [t-delta, t) ...
                if t - delta <= t_prev < t:
                    # ... and last node has to be e[0] ...
                    if e[0] in candidates[t_prev]:
                        for c in list(candidates[t_prev][e[0]]):

                            # c is path (p_0, p_1, ...) which ends in node e[0] at
                            # time t_prev
                            new_path = c + ((e[0], e[1], t),)

                            # we now know that (e[0], e[1]) is not the root of a
                            # new longest path as it continues a previous path c
                            root = False

                            # if c has previously been considered a longest path,
                            # we discard it from the list of longest paths. We also
                            # add the extended path as a new longest path
                            # (possible removing it later if it is further extended)
                            longest_paths.discard(c)
                            longest_paths.add(new_path)

                            # we add the newly found path as a candidate for paths
                            # which can be continued by future edges
                            if len(new_path) < max_length:
                                candidates[t][e[1]].add(new_path)

                            # if we delete candidate c we can only extend new_path from now on
                            # In this case, for every path we only consider the first link that 
                            # extends it to a longer path

                            candidates[t_prev][e[0]].discard(c)
            
            # if edge e does not continue a previous path
            # we start a new longest path
            if root:
                longest_paths.add(((e[0], e[1], t),))
                # add edge as candidate path of length one that can be extended by
                # future edges
                if max_length > 1:
                    candidates[t][e[1]].add(((e[0], e[1], t),))

        # we finished processing time stamp t, so
        # we can remove all candidates which finish
        # at a time smaller than t-delta. Since they cannot
        # be extended, these are longest paths
        for t_prev in list(candidates.keys()):
            if t_prev < t - delta:
                del candidates[t_prev]

    # once we reached the last time stamp, add all candidates
    # as longest paths
    # for t_prev in candidates:
    #    for x in candidates[t_prev]:
    #        for p in candidates[t_prev][x]:
    #            longest_paths.add(p)

    # Count occurrences as longest time-respecting path
    for x in longest_paths:
        path = (x[0][0],)
        for edge in x:
            path += (edge[1],)
        p.paths[len(x)][path][1] += 1

    # expand sub paths of longest paths
    p.expand_subpaths()

    Log.add('finished.')

    return p


def paths_from_temporal_network_dag(tempnet, delta=1, max_subpath_length=None):

    dag, node_map = DAG.from_temporal_network(tempnet, delta)
    return paths_from_dag(dag, node_map, max_subpath_length=max_subpath_length, repetitions=False)